"""

run -m fluidsim.base.forcing.milestone

solid_coarse -> solid

forcing(x, y, z) = sigma * solid(x, y) * (speed_target - velocity(x, y, z))

"""

from math import sin, cos, pi

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from fluiddyn.util import mpi

from fluidsim.operators.operators2d import OperatorsPseudoSpectral2D
from .specific import SpecificForcingPseudoSpectralSimple as Base


def step(x, limit, smoothness):
    return 0.5 * (np.tanh((-x + limit) / smoothness) + 1)


class PeriodicUniform:
    def __init__(self, speed_max, length, length_acc, lx):

        self.speed_max = speed_max

        sign = np.sign(speed_max)
        x_uni = length - 2 * length_acc
        t_uni = x_uni / abs(speed_max)
        self.acc = sign * speed_max ** 2 / (2 * length_acc)
        t_a = speed_max / self.acc
        self.period = 4 * t_a + 2 * t_uni

        self.t_1 = t_a
        self.t_2 = t_a + t_uni
        self.t_3 = self.t_2 + 2 * t_a
        self.t_4 = self.t_3 + t_uni

        self.x_0 = lx / 2 - sign * length / 2
        self.x_1 = self.x_4 = self.x_0 + sign * length_acc
        self.x_2 = self.x_3 = self.x_0 + sign * (length_acc + x_uni)

    def get_speed(self, time):

        speed_max = self.speed_max
        acc = self.acc
        t_1 = self.t_1
        t_2 = self.t_2
        t_3 = self.t_3
        t_4 = self.t_4

        time = time % self.period

        # acceleration
        if time <= t_1:
            return acc * time
        # uniform
        elif (time > t_1) and (time <= t_2):
            return speed_max
        # acceleration 2
        elif (time > t_2) and (time <= t_3):
            t = time - t_2
            return speed_max - acc * t
        # uniform 2
        elif (time > t_3) and (time <= t_4):
            return -speed_max
        # acceleration 3
        elif time > t_4:
            t = time - t_4
            return -speed_max + acc * t

    def get_locations(self, time):

        speed_max = self.speed_max
        acc = self.acc
        t_1 = self.t_1
        t_2 = self.t_2
        t_3 = self.t_3
        t_4 = self.t_4

        time = time % self.period

        # acceleration
        if time <= t_1:
            return self.x_0 + acc / 2 * time ** 2
        # uniform
        elif (time > t_1) and (time <= t_2):
            return self.x_1 + speed_max * (time - t_1)
        # acceleration 2
        elif (time > t_2) and (time <= t_3):
            t = time - t_2
            return self.x_2 + speed_max * t - acc / 2 * t ** 2
        # uniform 2
        elif (time > t_3) and (time <= t_4):
            return self.x_3 - speed_max * (time - t_3)
        # acceleration 3
        elif time > t_4:
            t = time - t_4
            return self.x_4 - speed_max * t + acc / 2 * t ** 2


class ForcingMilestone(Base):
    tag = "milestone"
    ndim = 2

    @classmethod
    def _complete_params_with_default(cls, params):
        Base._complete_params_with_default(params)
        milestone = params.forcing._set_child(cls.tag, dict(nx_max=None))
        milestone._set_child(
            "objects",
            dict(
                type="cylinders",
                number=2,
                diameter=1.0,
                width_boundary_layers=0.1,
            ),
        )
        movement = milestone._set_child("movement", dict(type="uniform"))
        movement._set_child("uniform", dict(speed=1.0))
        movement._set_child("sinusoidal", dict(length=1.0, period=1.0))
        movement._set_child(
            "periodic_uniform", dict(length=1.0, length_acc=1.0, speed=1.0)
        )

    @classmethod
    def get_info(cls, sim):
        info = {}
        params = sim.params
        params_milestone = params.forcing.milestone
        type_movement = params_milestone.movement.type
        if type_movement == "periodic_uniform":
            params_pu = params_milestone.movement.periodic_uniform
            periodic_uniform = PeriodicUniform(
                params_pu.speed,
                params_pu.length,
                params_pu.length_acc,
                params.oper.Lx,
            )
            info["period"] = periodic_uniform.period
        return info

    @classmethod
    def _modify_sim_repr_maker(cls, sim_repr_maker):
        sim = sim_repr_maker.sim
        params_milestone = sim.params.forcing.milestone
        type_movement = params_milestone.movement.type
        if type_movement == "periodic_uniform":
            params_pu = params_milestone.movement.periodic_uniform
            sim_repr_maker.add_parameters(
                {
                    "Lf": params_pu.length,
                    "U": params_pu.speed,
                    "D": params_milestone.objects.diameter,
                },
                formats={"Lf": ".2f"},
            )

    def _init_operators(self, sim):
        lx = sim.params.oper.Lx
        ly = sim.params.oper.Ly

        if self.params_milestone.nx_max is None:
            self._is_using_coarse_oper = False
            self.oper_coarse = sim.oper
        else:
            self._is_using_coarse_oper = True

            if mpi.rank == 0:
                params = OperatorsPseudoSpectral2D._create_default_params()
                params.oper.Lx = lx
                params.oper.Ly = ly
                nx = params.oper.nx = self.params_milestone.nx_max
                params.oper.ny = 2 * int(nx * ly / lx / 2)
                params.oper.type_fft = "sequential"
                self.oper_coarse = OperatorsPseudoSpectral2D(params)
                self.oper_coarse_shapeK_loc = self.oper_coarse.shapeK_loc
            else:
                self.oper_coarse = None
                self.oper_coarse_shapeK_loc = None

            if mpi.nb_proc > 1:
                self.oper_coarse_shapeK_loc = mpi.comm.bcast(
                    self.oper_coarse_shapeK_loc, root=0
                )

            self.solid = self.sim.oper.create_arrayX(value=0)
            self.solid_fft = self.sim.oper.create_arrayK(value=0)

    def __init__(self, sim):
        super().__init__(sim)

        self.params_milestone = sim.params.forcing.milestone

        nx_max = self.params_milestone.nx_max

        if nx_max is not None and nx_max > sim.params.oper.nx:
            raise ValueError("params_milestone.nx_max > sim.params.oper.nx")

        if nx_max == sim.params.oper.nx:
            self.params_milestone.nx_max = None

        self._init_operators(sim)

        # Calculus of coef_sigma:
        # f(t) = f0 * exp(-sigma*t)
        # If we want f(t)/f0 = 10**(-gamma) after n_dt time steps, we have to have:
        # sigma = gamma / (n_dt * dt)
        gamma = 2
        n_dt = 4
        self.coef_sigma = gamma / n_dt
        self.sigma = self.coef_sigma / self.params.time_stepping.deltat_max

        self.number_objects = self.params_milestone.objects.number
        mesh = sim.params.oper.Ly / self.number_objects
        self.y_coors = mesh * (1 / 2 + np.arange(self.number_objects))

        type_movement = self.params_milestone.movement.type
        if type_movement == "uniform":
            self.get_locations = self.get_locations_uniform
            self._speed = self.params_milestone.movement.uniform.speed
            self.get_speed = self.get_speed_uniform
        elif type_movement == "sinusoidal":
            sinusoidal = self.params_milestone.movement.sinusoidal
            self._half_length = sinusoidal.length / 2
            self._omega = 2 * pi / sinusoidal.period
            self.get_locations = self.get_locations_sinusoidal
            self.get_speed = self.get_speed_sinusoidal

        elif type_movement == "periodic_uniform":

            params_pu = self.params_milestone.movement.periodic_uniform
            periodic_uniform = PeriodicUniform(
                params_pu.speed,
                params_pu.length,
                params_pu.length_acc,
                sim.params.oper.Lx,
            )

            def get_locations(time):
                x_coors = periodic_uniform.get_locations(time) * np.ones(
                    self.number_objects
                )
                return x_coors, self.y_coors

            self.get_locations = get_locations
            self.get_speed = periodic_uniform.get_speed
            self.period = periodic_uniform.period
        else:
            raise NotImplementedError

    def get_solid_field(self, time):

        if mpi.rank > 0 and (self._is_using_coarse_oper or self.ndim == 3):
            return (None,) * 3

        oper = self.oper_coarse
        lx = oper.Lx
        radius = self.params_milestone.objects.diameter / 2

        width = self.params_milestone.objects.width_boundary_layers

        x_coors, y_coors = self.get_locations(time)
        solid = np.zeros_like(oper.X)
        for x, y in zip(x_coors, y_coors):
            for index_x_periodicity in range(-1, 2):
                x_center = x + index_x_periodicity * lx
                distance_from_center = np.sqrt(
                    (oper.X - x_center) ** 2 + (oper.Y - y) ** 2
                )
                solid += step(distance_from_center, radius, width)
        return solid, x_coors, y_coors

    def get_locations_uniform(self, time):
        speed = self.params_milestone.movement.uniform.speed
        lx = self.params.oper.Lx
        x_coors = (speed * time) % lx * np.ones(self.number_objects)
        return x_coors, self.y_coors

    def get_speed_uniform(self, time):
        return self._speed

    def get_locations_sinusoidal(self, time):
        lx = self.params.oper.Lx
        x_coors = (
            lx / 2 + self._half_length * sin(self._omega * time - pi / 2)
        ) * np.ones(self.number_objects)
        return x_coors, self.y_coors

    def get_speed_sinusoidal(self, time):
        return self._half_length * self._omega * cos(self._omega * time - pi / 2)

    def check_plot_solid(self, time):

        solid, x_coors, y_coors = self.get_solid_field(time)

        fig, ax = plt.subplots()
        oper_c = self.oper_coarse
        lx = oper_c.Lx
        ly = oper_c.Ly

        nx = oper_c.nx
        ny = oper_c.ny
        xs = np.linspace(0, lx, nx + 1)
        ys = np.linspace(0, ly, ny + 1)
        pcmesh = ax.pcolormesh(xs, ys, solid)
        ax.axis("equal")
        ax.set_xlim((0, lx))
        ax.set_ylim((0, ly))
        fig.colorbar(pcmesh)

    def check_plot_forcing(self, time):

        self.compute(time)

        try:
            fx = self.fstate.state_phys.get_var("ux")
        except ValueError:
            raise NotImplementedError(
                "check_plot_forcing is not implemented for this solver. "
                "You can try check_plot_solid or check_with_animation."
            )

        fy = self.fstate.state_phys.get_var("uy")

        rot_f = self.fstate.state_phys.get_var("rot")

        fig, ax = plt.subplots()
        oper = self.sim.oper
        lx = oper.Lx
        ly = oper.Ly

        nx = oper.nx
        ny = oper.ny
        xs = np.linspace(0, lx, nx + 1)
        ys = np.linspace(0, ly, ny + 1)
        pcmesh = ax.pcolormesh(xs, ys, rot_f)

        ax.quiver(oper.X, oper.Y, fx, fy)

        ax.axis("equal")
        ax.set_xlim((0, lx))
        ax.set_ylim((0, ly))
        fig.colorbar(pcmesh)

    def check_with_animation(self, number_frames=40, interval=500):

        oper_c = self.oper_coarse

        lx = oper_c.Lx
        ly = oper_c.Ly

        nx = oper_c.nx
        ny = oper_c.ny

        movement = self.params_milestone.movement
        type_movement = movement.type

        try:
            period = self.period
        except AttributeError:
            if type_movement == "uniform":
                period = lx / movement.uniform.speed
            elif type_movement == "sinusoidal":
                period = movement.sinusoidal.period
            else:
                raise NotImplementedError

        dt = period / number_frames

        fig, ax = plt.subplots()

        solid, x_coors, y_coors = self.get_solid_field(0)

        scat = ax.scatter(x_coors, y_coors, marker="x")
        ax.axis("equal")

        xs = np.linspace(0, lx, nx + 1)
        ys = np.linspace(0, ly, ny + 1)
        pcmesh = ax.pcolormesh(xs, ys, solid)

        ax.set_xlim((0, lx))
        ax.set_ylim((0, ly))
        fig.colorbar(pcmesh)

        def update_plot(index_time, pcmesh, scat):
            time = index_time * dt
            solid, x_coors, y_coors = self.get_solid_field(time)
            scat.set_offsets(np.vstack((x_coors, y_coors)).T)
            pcmesh.set_array(solid.flatten())
            return (pcmesh, scat)

        self._animation = animation.FuncAnimation(
            fig,
            update_plot,
            frames=range(number_frames),
            fargs=(pcmesh, scat),
            blit=True,
            interval=interval,  # in ms
        )
        plt.show()

    def _full_from_coarse(self, solid):
        if mpi.rank == 0:
            solid_coarse_fft = self.oper_coarse.fft(solid)
            nKyc, nKxc = self.oper_coarse_shapeK_loc
            solid_coarse_fft[nKyc // 2, :] = 0.0
            solid_coarse_fft[:, nKxc - 1] = 0.0
        else:
            solid_coarse_fft = None

        self.sim.oper.put_coarse_array_in_array_fft(
            solid_coarse_fft,
            self.solid_fft,
            self.oper_coarse,
            self.oper_coarse_shapeK_loc,
        )

        self.sim.oper.ifft_as_arg(self.solid_fft, self.solid)
        return self.solid

    def compute(self, time=None):

        sim = self.sim

        if time is None:
            time = sim.time_stepping.t

        solid, x_coors, y_coors = self.get_solid_field(time)

        if self._is_using_coarse_oper:
            solid = self._full_from_coarse(solid)

        ux = sim.state.state_phys.get_var("ux")
        fx = self.sigma * solid * (self.get_speed(time) - ux)
        fx_fft = sim.oper.fft(fx)

        if "rot_fft" in sim.state.keys_state_spect:
            fy_fft = sim.oper.create_arrayK(value=0)
            rot_fft = self.oper.rotfft_from_vecfft(fx_fft, fy_fft)
            self.fstate.init_statespect_from(rot_fft=rot_fft)
        else:
            uy = sim.state.state_phys.get_var("uy")
            fy = -self.sigma * solid * uy
            fy_fft = sim.oper.fft(fy)
            if sim.params.oper.NO_SHEAR_MODES:
                sim.oper.dealiasing(fx_fft, fy_fft)
            self.fstate.init_statespect_from(ux_fft=fx_fft, uy_fft=fy_fft)


if __name__ == "__main__":

    # from time import perf_counter

    from fluidsim.solvers.ns2d.with_uxuy import Simul

    # from fluidsim.solvers.ns2d.solver import Simul

    params = Simul.create_default_params()

    diameter = 0.5
    number_cylinders = 3
    speed = 0.1

    ny = params.oper.ny = 96 * 2
    nx = params.oper.nx = 4 * ny / 3

    ly = params.oper.Ly = 3 * diameter * number_cylinders
    lx = params.oper.Lx = ly / ny * nx * 4 / number_cylinders

    params.forcing.enable = True
    params.forcing.type = "milestone"
    params.forcing.milestone.nx_max = 48
    objects = params.forcing.milestone.objects

    objects.number = number_cylinders
    objects.diameter = diameter
    objects.width_boundary_layers = 0.1

    movement = params.forcing.milestone.movement

    # movement.type = "uniform"
    # movement.uniform.speed = 1.0

    # movement.type = "sinusoidal"
    # movement.sinusoidal.length = 14 * diameter
    # movement.sinusoidal.period = 100.0

    movement.type = "periodic_uniform"
    movement.periodic_uniform.length = 14 * diameter
    movement.periodic_uniform.length_acc = 0.25
    movement.periodic_uniform.speed = speed

    params.init_fields.type = "noise"
    params.init_fields.noise.velo_max = 5e-3

    movement = PeriodicUniform(
        speed,
        movement.periodic_uniform.length,
        movement.periodic_uniform.length_acc,
        lx,
    )

    params.time_stepping.t_end = movement.period * 2

    params.nu_8 = 1e-14

    params.output.periods_plot.phys_fields = 5.0
    params.output.periods_print.print_stdout = 5

    sim = Simul(params)

    if mpi.rank == 0 and params.output.periods_plot.phys_fields:
        fig = plt.gcf()
        ax = fig.axes[0]
        ax.axis("equal")
        fig.set_size_inches(14, 4)

    self = milestone = sim.forcing.forcing_maker

    # t_start = perf_counter()
    # sim.time_stepping.start()
    # mpi.printby0(f"Done in {perf_counter() - t_start} s")

    # milestone.check_with_animation()
    # milestone.check_plot_solid(0)
    # self.check_plot_forcing(10)



from fluiddyn.util.compat import cached_property
from fluidsim.util.testing import TestSimul, skip_if_no_fluidfft, classproperty


@skip_if_no_fluidfft
class TestSimulBase(TestSimul):
    @classproperty
    def Simul(cls):
        from .solver import Simul

        return Simul

    @classmethod
    def init_params(cls):
        params = cls.params = cls.Simul.create_default_params()

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"

        nh = 32
        lx = 2.0
        params.oper.nx = nh
        params.oper.ny = nh
        params.oper.Lx = lx
        params.oper.Ly = lx

        params.oper.coef_dealiasing = 2.0 / 3
        params.nu_8 = 2.0

        params.time_stepping.t_end = 0.5
        params.time_stepping.USE_CFL = False

        return params


class TestOutput(TestSimulBase):
    @classmethod
    def init_params(cls):
        params = super().init_params()

        params.init_fields.type = "noise"

        params.output.periods_save.phys_fields = 0.5
        params.output.periods_plot.phys_fields = 0.1

        params.output.ONLINE_PLOT_OK = True
        params.output.phys_fields.field_to_plot = "f"
        params.output.periods_print.print_stdout = 0.1

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.2

    def test_output(self):
        sim = self.sim
        sim.time_stepping.start()



"""Stratified Navier-Stokes 2D (:mod:`fluidsim.solvers.ns2d.strat`)
===================================================================

Provides:

.. autosummary::
   :toctree:

   solver
   state
   init_fields
   output
   forcing

"""



import h5py
import os
import numpy as np

from transonic import boost, Array
from fluiddyn.util import mpi

from .base import SpecificOutput

Ai = Array[np.int32, "1d"]
Af = Array[float, "2d"]


@boost
def strfunc_from_pdf(
    rxs: Ai, pdf: Af, values: Af, order: float, absolute: bool = False
):
    """Compute structure function of specified order from pdf for increments
    module.

    """
    S_order = np.empty(rxs.shape)
    if absolute:
        values = abs(values)
    for irx in range(rxs.size):
        deltainc = abs(values[irx, 1] - values[irx, 0])
        S_order[irx] = deltainc * np.sum(pdf[irx] * values[irx] ** order)

    return S_order


class Increments(SpecificOutput):
    """A :class:`Increments` object handles the saving of pdf of
    increments.
    """

    _tag = "increments"
    _name_file = _tag + ".h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "increments"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(tag, attribs={"HAS_TO_PLOT_SAVED": False})

    def __init__(self, output):
        params = output.sim.params
        self.nx = params.oper.nx

        self.nrx = min(self.nx // 16, 128)
        self.nrx = int(max(self.nrx, self.nx // 2))
        rmin = 1
        rmax = int(0.8 * self.nx)
        delta_logr = np.log(rmax / rmin) / (self.nrx - 1)
        logr = np.log(rmin) + delta_logr * np.arange(self.nrx)
        self.rxs = np.array(np.round(np.exp(logr)), dtype=np.int32)

        for ir in range(1, self.nrx):
            if self.rxs[ir - 1] >= self.rxs[ir]:
                self.rxs[ir] = self.rxs[ir - 1] + 1

        self.nbins = 400

        self.output = output
        self._init_path_files()

        if os.path.exists(self.path_file):
            if mpi.rank == 0:
                with h5py.File(self.path_file, "r") as h5file:
                    self.rxs = h5file["rxs"][...]
                    self.nbins = h5file["nbins"][...]
            if mpi.nb_proc > 1:
                self.rxs = mpi.comm.bcast(self.rxs)
                self.nbins = mpi.comm.bcast(self.nbins)

        self.nrx = self.rxs.size
        arrays_1st_time = {"rxs": self.rxs, "nbins": self.nbins}
        self._bins = np.arange(0.5, self.nbins, dtype=float) / self.nbins
        self.keys_vars_to_compute = list(output.sim.state.state_phys.keys)

        super().__init__(
            output,
            period_save=params.output.periods_save.increments,
            has_to_plot_saved=params.output.increments.HAS_TO_PLOT_SAVED,
            arrays_1st_time=arrays_1st_time,
        )

    def _init_online_plot(self):
        if mpi.rank == 0:
            self.fig, axe = self.output.figure_axe(numfig=5_000_000)
            self.axe = axe
            axe.set_xlabel(r"$\delta u_x (x)$")
            axe.set_ylabel("pdf")
            axe.set_title(
                r"pdf $\delta u_x (x)$" + "\n" + self.output.summary_simul
            )

    def _online_plot_saving(self, dict_results, key="rot"):
        """online plot on pdf"""
        pdf = dict_results["pdf_delta_" + key]
        pdf = pdf.reshape([self.nrx, self.nbins])
        valmin = dict_results["valmin_" + key]
        valmax = dict_results["valmax_" + key]

        for irx, rx in enumerate(self.rxs):
            values_inc = self.compute_values_inc(valmin[irx], valmax[irx])
            self.axe.plot(values_inc + irx, pdf[irx])

    def compute(self):
        """compute the values at one time."""
        dict_results = {}
        for key in self.keys_vars_to_compute:
            var = self.sim.state.get_var(key)

            pdf_var = np.empty([self.nrx, self.nbins])
            valmin = np.empty([self.nrx])
            valmax = np.empty([self.nrx])

            for irx, rx in enumerate(self.rxs):
                inc_var = self.oper.compute_increments_dim1(var, rx)
                (pdf_var[irx], bin_edges_var) = self.oper.pdf_normalized(
                    inc_var, self.nbins
                )
                valmin[irx] = bin_edges_var[0]
                valmax[irx] = bin_edges_var[self.nbins]

            dict_results["pdf_delta_" + key] = pdf_var.flatten()
            dict_results["valmin_" + key] = valmin
            dict_results["valmax_" + key] = valmax

        return dict_results

    def compute_values_inc(self, valmin, valmax):
        return valmin + (valmax - valmin) * self._bins

    def load(self):
        """load the saved pdf and return a dictionary."""
        with h5py.File(self.path_file, "r") as h5file:
            times = h5file["times"][...]

            list_base_keys = [
                "pdf_delta_",
                "valmin_",
                "valmax_",
                # 'struc_func_'
            ]

            dict_results = {"times": times}
            for key in self.keys_vars_to_compute:
                for base_key in list_base_keys:
                    result = h5file[base_key + key][...]
                    dict_results[base_key + key] = result

        return dict_results

    def plot(self, tmin=0, tmax=None, delta_t=2, order=2, yscale="log"):
        """Plot some structure functions."""
        with h5py.File(self.path_file, "r") as h5file:
            times = h5file["times"][...]
            # nt = len(times)

            if tmax is None:
                tmax = times.max()

            rxs = h5file["rxs"][...]

            oper = h5file["/info_simul/params/oper"]
            nx = oper.attrs["nx"]
            Lx = oper.attrs["Lx"]
        deltax = Lx / nx

        rxs = np.array(rxs, dtype=np.float64) * deltax

        # orders = h5file['orders'][...]
        # dset_struc_func_ux = h5file['struc_func_ux']
        # dset_struc_func_uy = h5file['struc_func_uy']

        delta_t_save = np.mean(times[1:] - times[0:-1])
        delta_i_plot = int(np.round(delta_t / delta_t_save))
        if delta_i_plot == 0 and delta_t != 0.0:
            delta_i_plot = 1
        delta_t = delta_i_plot * delta_t_save

        imin_plot = np.argmin(abs(times - tmin))
        imax_plot = np.argmin(abs(times - tmax))

        tmin_plot = times[imin_plot]
        tmax_plot = times[imax_plot]

        to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
            tmin, tmax, delta_t
        )
        print(to_print)

        to_print = (
            "plot structure functions\n"
            "tmin = {0:8.6g} ; tmax = {1:8.6g} ; delta_t = {2:8.6g}\n"
            "imin = {3:8d} ; imax = {4:8d} ; delta_i = {5:8d}"
        ).format(
            tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
        )
        print(to_print)

        pdf_ux, values_inc_ux, nb_rx_to_plot = self.load_pdf_from_file(
            tmin=tmin, tmax=tmax, key_var="ux"
        )

        pdf_uy, values_inc_uy, nb_rx_to_plot = self.load_pdf_from_file(
            tmin=tmin, tmax=tmax, key_var="uy"
        )

        # iorder = self.iorder_from_order(order)
        order = float(order)

        x_left_axe = 0.12
        z_bottom_axe = 0.56
        width_axe = 0.85
        height_axe = 0.37
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("$r_x$")
        ax1.set_ylabel(r"$\langle \delta u^{" + f"{order}" + "} \\rangle$")

        ax1.set_title("struct. functions\n" + self.output.summary_simul)
        ax1.set_xscale("log")
        ax1.set_yscale(yscale)

        So_ux = self.strfunc_from_pdf(pdf_ux, values_inc_ux, order)
        So_uy = self.strfunc_from_pdf(pdf_uy, values_inc_uy, order)

        norm = rxs

        # ax1.set_ylabel('struct. functions, order = {0}'.format(order))
        # if delta_t != 0.:
        #     for it in range(imin_plot,imax_plot+1,delta_i_plot):
        #         struc_func_ux = dset_struc_func_ux[it]
        #         struc_func_ux = struc_func_ux.reshape(
        #             [self.norders, self.nrx])
        #         struc_func_uy = dset_struc_func_uy[it]
        #         struc_func_uy = struc_func_uy.reshape(
        #             [self.norders, self.nrx])

        #         ax1.plot(rxs, struc_func_ux[iorder], 'c', linewidth=1)
        #         ax1.plot(rxs, struc_func_uy[iorder], 'm', linewidth=1)

        # struc_func_ux = dset_struc_func_ux[imin_plot:imax_plot+1].mean(0)
        # struc_func_ux = struc_func_ux.reshape([self.norders, self.nrx])
        # struc_func_uy = dset_struc_func_uy[imin_plot:imax_plot+1].mean(0)
        # struc_func_uy = struc_func_uy.reshape([self.norders, self.nrx])

        # ax1.plot(rxs, struc_func_ux[iorder]/norm, 'c', linewidth=2)
        # ax1.plot(rxs, struc_func_uy[iorder]/norm, 'm', linewidth=2)

        ax1.plot(rxs, So_ux / norm, "c-.", linewidth=2)
        ax1.plot(rxs, So_uy / norm, "m-.", linewidth=2)
        if order % 2 == 1:
            ax1.plot(rxs, -So_ux / norm, "c:", linewidth=2)
            ax1.plot(rxs, -So_uy / norm, "m:", linewidth=2)

        # ax1.plot(rxs, abs(struc_func_ux[iorder])/abs(struc_func_uy[iorder]),
        #          'k', linewidth=1)

        ax1.plot(rxs, abs(So_ux) / abs(So_uy), "k", linewidth=1)

        # if self.orders[iorder]%2 == 1:
        #     ax1.plot(rxs, -struc_func_ux[iorder]/norm, '--b', linewidth=2)
        #     ax1.plot(rxs, -struc_func_uy[iorder]/norm, '--m', linewidth=2)

        cond = rxs < 6 * deltax
        ax1.plot(
            rxs[cond], 1.0e4 * rxs[cond] ** (order) / norm[cond], "k", linewidth=2
        )
        ax1.plot(rxs, rxs ** (order / 3) / norm, "--k", linewidth=2)

        ax1.plot(rxs, 1.0e0 * rxs ** (1) / norm, ":k", linewidth=2)

        z_bottom_axe = 0.09
        size_axe[1] = z_bottom_axe
        ax2 = fig.add_axes(size_axe)

        ax2.set_xlabel("$r_x$")
        ax2.set_ylabel("flatness")
        ax2.set_xscale("log")
        ax2.set_yscale("log")

        # iorder4 = self.iorder_from_order(4)
        # iorder2 = self.iorder_from_order(2)

        # if delta_t != 0.:
        #     for it in range(imin_plot,imax_plot+1,delta_i_plot):
        #         struc_func_ux = dset_struc_func_ux[it]
        #         struc_func_ux = struc_func_ux.reshape(
        #             [self.norders, self.nrx])
        #         struc_func_uy = dset_struc_func_uy[it]
        #         struc_func_uy = struc_func_uy.reshape(
        #             [self.norders, self.nrx])

        #         flatnessL = struc_func_ux[iorder4]/struc_func_ux[iorder2]**2
        #         flatnessT = struc_func_uy[iorder4]/struc_func_uy[iorder2]**2

        #         ax2.plot(rxs, flatnessL, 'c', linewidth=1)
        #         ax2.plot(rxs, flatnessT, 'm', linewidth=1)

        # struc_func_ux = dset_struc_func_ux[imin_plot:imax_plot+1].mean(0)
        # struc_func_ux = struc_func_ux.reshape([self.norders, self.nrx])
        # struc_func_uy = dset_struc_func_uy[imin_plot:imax_plot+1].mean(0)
        # struc_func_uy = struc_func_uy.reshape([self.norders, self.nrx])

        # flatnessL = struc_func_ux[iorder4]/struc_func_ux[iorder2]**2
        # flatnessT = struc_func_uy[iorder4]/struc_func_uy[iorder2]**2
        # ax2.plot(rxs, flatnessL, 'c', linewidth=2)
        # ax2.plot(rxs, flatnessT, 'm', linewidth=2)

        S2_ux = self.strfunc_from_pdf(pdf_ux, values_inc_ux, 2)
        S2_uy = self.strfunc_from_pdf(pdf_uy, values_inc_uy, 2)

        S4_ux = self.strfunc_from_pdf(pdf_ux, values_inc_ux, 4)
        S4_uy = self.strfunc_from_pdf(pdf_uy, values_inc_uy, 4)

        flatnessL_bis = S4_ux / S2_ux ** 2
        flatnessT_bis = S4_uy / S2_uy ** 2

        ax2.plot(rxs, flatnessL_bis, "c--", linewidth=2)
        ax2.plot(rxs, flatnessT_bis, "m--", linewidth=2)

        cond = np.logical_and(rxs < 70 * deltax, rxs > 5 * deltax)
        ax2.plot(rxs[cond], 1e1 * rxs[cond] ** (-1), ":k", linewidth=2)

        ax2.plot(rxs, 3 * np.ones(rxs.shape), "k--", linewidth=0.5)

    def strfunc_from_pdf(self, pdf, values, order, absolute=False):
        r"""Following the identity:
        .. math::
            E(x^m) = \int_{-\inf}^{\inf} x^m p(x) dx

        In this case, replace x with increments,
        .. math::
            \delta u(r, x) = u(x+r) - u(x)

        Thus, for a every value of r the mean of increments are computed
        as follows:
        .. math::
            <(\delta u)^m>
                = \int_{-\inf}^{\inf} (\delta u)^m p(\delta u) d(\delta u)
                = d(\delta u) \Sigma (\delta u)^m p(\delta u)
        """
        return strfunc_from_pdf(self.rxs, pdf, values, float(order), absolute)

    def load_pdf_from_file(
        self, tmin=0, tmax=None, key_var="ux", irx_to_plot=None
    ):
        """Plot some pdf."""
        with h5py.File(self.path_file, "r") as h5file:
            times = h5file["times"][...]
            nt = len(times)

            if tmax is None:
                tmax = times.max()

            rxs = h5file["rxs"][...]

            oper = h5file["/info_simul/params/oper"]
            nx = oper.attrs["nx"]
            Lx = oper.attrs["Lx"]

            deltax = Lx / nx

            rxs = np.array(rxs, dtype=np.float64) * deltax

            # orders = h5file['orders'][...]

            # delta_t_save = np.mean(times[1:]-times[0:-1])
            # delta_t = delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            # tmin_plot = times[imin_plot]
            # tmax_plot = times[imax_plot]

            #         to_print = '''load pdf of the increments
            # tmin = {0:8.6g} ; tmax = {1:8.6g}
            # imin = {2:8d} ; imax = {3:8d}'''.format(
            # tmin_plot, tmax_plot,
            # imin_plot, imax_plot)
            #         print(to_print)

            if irx_to_plot is None:
                irx_to_plot = np.arange(rxs.size)

            nb_rx_to_plot = irx_to_plot.size

            # print 'irx_to_plot', irx_to_plot
            # print 'self.rxs[irx_to_plot]', self.rxs[irx_to_plot]

            pdf_timemean = np.zeros([nb_rx_to_plot, self.nbins])
            values_inc_timemean = np.zeros([nb_rx_to_plot, self.nbins])

            valmin_timemean = np.zeros([nb_rx_to_plot])
            valmax_timemean = np.zeros([nb_rx_to_plot])
            nb_timemean = 0

            for it in range(imin_plot, imax_plot + 1):
                nb_timemean += 1
                valmin = h5file["valmin_" + key_var][it]
                valmax = h5file["valmax_" + key_var][it]

                for irxp, irx in enumerate(irx_to_plot):
                    valmin_timemean[irxp] += valmin[irx]
                    valmax_timemean[irxp] += valmax[irx]

            valmin_timemean /= nb_timemean
            valmax_timemean /= nb_timemean

            for irxp, irx in enumerate(irx_to_plot):
                values_inc_timemean[irxp] = self.compute_values_inc(
                    valmin_timemean[irxp], valmax_timemean[irxp]
                )

            nt = 0
            for it in range(imin_plot, imax_plot + 1):
                nt += 1
                pdf_dvar2D = h5file["pdf_delta_" + key_var][it]
                pdf_dvar2D = pdf_dvar2D.reshape([self.nrx, self.nbins])
                valmin = h5file["valmin_" + key_var][it]
                valmax = h5file["valmax_" + key_var][it]

                for irxp, irx in enumerate(irx_to_plot):
                    pdf_dvar = pdf_dvar2D[irx]
                    values_inc = self.compute_values_inc(valmin[irx], valmax[irx])

                    pdf_timemean[irxp] += np.interp(
                        values_inc_timemean[irxp], values_inc, pdf_dvar
                    )

        pdf_timemean /= nt

        return pdf_timemean, values_inc_timemean, nb_rx_to_plot

    def plot_pdf(self, tmin=0, tmax=None, key_var="ux", order=0, nb_rx_to_plot=5):

        irx_to_plot = np.arange(
            0, self.rxs.size, self.rxs.size / nb_rx_to_plot
        ).astype(int)
        nb_rx_to_plot = irx_to_plot.size

        (
            pdf_timemean,
            values_inc_timemean,
            nb_rx_to_plot,
        ) = self.load_pdf_from_file(
            tmin=tmin, tmax=tmax, key_var=key_var, irx_to_plot=irx_to_plot
        )

        to_print = f"plot_pdf(tmin={tmin}, tmax={tmax})"
        print(to_print)

        fig, ax1 = self.output.figure_axe()
        ax1.set_title("pdf increments\n" + self.output.summary_simul)

        ax1.set_xscale("linear")
        ax1.set_yscale("linear")

        ax1.set_xlabel(key_var)
        ax1.set_ylabel(r"PDF x $\delta v^" + repr(order) + "$")

        colors = ["k", "y", "r", "b", "g", "m", "c"]

        for irxp, irx in enumerate(irx_to_plot):

            print("color = {}, rx = {}".format(colors[irxp], self.rxs[irx]))

            val_inc = values_inc_timemean[irxp]

            ax1.plot(
                val_inc,
                pdf_timemean[irxp] * val_inc ** order,
                colors[irxp] + "x-",
                linewidth=1,
            )


# def iorder_from_order(self, order):
#     """Return the indice corresponding to one value of order."""
#     iorder = abs(self.orders-order).argmin()
#     if self.orders[iorder] != order:
#         raise ValueError(
#             'Order {0} has not been computed ?'.format(order)
#             )
#     return iorder



"""
Plate2d output (:mod:`fluidsim.solvers.plate2d.output`)
=============================================================


Provides:

.. autosummary::
   :toctree:

   print_stdout
   spatial_means
   spectra
   correlations_freq

"""

import numpy as np


from fluidsim.base.output import OutputBasePseudoSpectral


class Output(OutputBasePseudoSpectral):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        classes = info_solver.classes.Output._set_child("classes")

        package = "fluidsim.solvers.plate2d.output"

        classes._set_child(
            "PrintStdOut",
            attribs={
                "module_name": package + ".print_stdout",
                "class_name": "PrintStdOutPlate2D",
            },
        )

        classes._set_child(
            "PhysFields",
            attribs={
                "module_name": "fluidsim.base.output.phys_fields2d",
                "class_name": "PhysFieldsBase2D",
            },
        )

        classes._set_child(
            "Spectra",
            attribs={
                "module_name": package + ".spectra",
                "class_name": "SpectraPlate2D",
            },
        )

        classes._set_child(
            "spatial_means",
            attribs={
                "module_name": package + ".spatial_means",
                "class_name": "SpatialMeansPlate2D",
            },
        )

        classes._set_child(
            "correl_freq",
            attribs={
                "module_name": package + ".correlations_freq",
                "class_name": "CorrelationsFreq",
            },
        )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        OutputBasePseudoSpectral._complete_params_with_default(
            params, info_solver
        )

        params.output.phys_fields.field_to_plot = "z"

    def compute_energies_conversion_fft(self):
        w_fft = self.sim.state.get_var("w_fft")
        z_fft = self.sim.state.get_var("z_fft")
        chi_fft = self.sim.state.get_var("chi_fft")
        K4 = self.sim.oper.K4

        Ek_fft = 0.5 * np.abs(w_fft) ** 2
        El_fft = np.abs(0.5 * K4 * np.abs(z_fft) ** 2)
        Ee_fft = np.abs(
            0.25 * self.sim.oper.laplacian_fft(np.abs(chi_fft) ** 2 + 0j, order=4)
        )

        conversion_k_to_l_fft = np.real(K4 * w_fft * z_fft.conj())

        mamp_wz = self.sim.oper.monge_ampere_from_fft(w_fft, z_fft)
        conversion_l_to_e_fft = -np.real(
            self.sim.oper.fft2(mamp_wz) * chi_fft.conj()
        )

        return (
            Ek_fft,
            El_fft,
            Ee_fft,
            conversion_k_to_l_fft,
            conversion_l_to_e_fft,
        )

    def compute_energies_fft(self):
        w_fft = self.sim.state.state_spect.get_var("w_fft")
        z_fft = self.sim.state.state_spect.get_var("z_fft")
        chi_fft = self.sim.state.get_var("chi_fft")
        Ek_fft = 0.5 * np.abs(w_fft) ** 2
        El_fft = np.abs(
            0.5 * self.sim.oper.laplacian_fft(np.abs(z_fft) ** 2 + 0j, order=4)
        )
        Ee_fft = np.abs(
            0.25 * self.sim.oper.laplacian_fft(np.abs(chi_fft) ** 2 + 0j, order=4)
        )
        return Ek_fft, El_fft, Ee_fft

    def compute_energy_fft(self):
        Ek_fft, El_fft, Ee_fft = self.compute_energies_fft()
        return Ek_fft + El_fft + Ee_fft

    def compute_energy(self):
        E_fft = self.compute_energy_fft()
        return self.sum_wavenumbers(E_fft)



"""AD1D solver (:mod:`fluidsim.solvers.ad1d.solver`)
====================================================

Provides:

.. autoclass:: Simul
   :members:
   :private-members:

"""

from fluidsim.base.solvers.base import SimulBase
from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.finite_diff import InfoSolverFiniteDiff


class InfoSolverAD1D(InfoSolverFiniteDiff):
    def _init_root(self):

        super()._init_root()

        package = "fluidsim.solvers.ad1d"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "AD1D"

        classes = self.classes

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateAD1D"

        classes.InitFields.module_name = package + ".init_fields"
        classes.InitFields.class_name = "InitFieldsAD1D"

        classes.Output.module_name = package + ".output"
        classes.Output.class_name = "Output"


class Simul(SimulBase):
    r"""Advection-diffusion solver 1D.

    Notes
    -----

    .. |p| mathmacro:: \partial

    We use a finite difference method with the Crank-Nicolson time
    scheme to solve the equation

    .. math:: \p_t s + U \p_x s = D(s),

    where :math:`d(s)` is the dissipation term and :math:`U` is a
    constant velocity.

    """
    InfoSolver = InfoSolverAD1D

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        SimulBase._complete_params_with_default(params)
        attribs = {"U": 1.0}
        params._set_attribs(attribs)

    def tendencies_nonlin(self, state_phys=None, old=None):
        """Compute the "nonlinear" tendencies."""
        if old is None:
            tendencies = SetOfVariables(
                like=self.state.state_phys, info="tendencies", value=0.0
            )
        else:
            tendencies = old

        if self.params.forcing.enable:
            tendencies += self.forcing.tendencies

        return tendencies

    def linear_operator(self):
        """Compute the linear operator as a matrix."""

        return (
            self.params.nu_2 * (self.oper.sparse_pxx)
            - self.params.U * self.oper.sparse_px
        )


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.U = 1.0

    params.short_name_type_run = "test"

    params.oper.nx = 200
    params.oper.Lx = 1.0

    # params.oper.type_fft = 'FFTWPY'

    params.time_stepping.type_time_scheme = "RK2"

    # delta_x = params.oper.Lx/params.oper.nx
    params.nu_2 = 0.01

    params.time_stepping.t_end = 0.4
    params.time_stepping.USE_CFL = True
    # params.time_stepping.deltat0 = 0.1

    params.init_fields.type = "gaussian"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 0.5

    params.output.periods_plot.phys_fields = 0.0

    params.output.phys_fields.field_to_plot = "s"

    # params.output.spectra.has_to_plot = 1  # False
    # params.output.spatial_means.has_to_plot = 1  # False
    # params.output.spect_energy_budg.has_to_plot = 1  # False
    # params.output.increments.has_to_plot = 1  # False

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()

    print("x of s_max: ", sim.oper.xs[sim.state.state_phys.argmax()])

    sim.output.phys_fields.plot()

    fld.show()



"""Simple text output (:mod:`fluidsim.solvers.ns2d.strat.output.print_stdout`)
==============================================================================

.. autoclass:: PrintStdOutNS2DStrat
   :members:
   :private-members:

"""

import os

import numpy as np

from fluidsim.base.output.print_stdout import PrintStdOutBase

from fluiddyn.util import get_memory_usage
from fluiddyn.util import mpi


class PrintStdOutNS2DStrat(PrintStdOutBase):
    """Simple text output.

    Used to print in both the stdout and the stdout.txt file, and also
    to print simple info on the current state of the simulation.

    """

    def __init__(self, output):
        super().__init__(output)
        self.path_memory = self.output.path_run + "/memory_out.txt"
        if mpi.rank == 0 and self.output._has_to_save:
            if not os.path.exists(self.path_memory):
                self.file_memory = open(self.path_memory, "w")
            else:
                self.file_memory = open(self.path_memory, "r+")
                self.file_memory.seek(0, 2)  # go to the end of the file

    def _make_str_info(self):
        to_print = super()._make_str_info()

        energyK, energyA, energyK_ux = self.output.compute_energies()
        energy = energyK + energyA
        if mpi.rank == 0:
            to_print += (
                "              energyK = {:9.3e}\n"
                "              energyA = {:9.3e}\n"
                "              energy  = {:9.3e} ; Delta energy = {:+9.3e}\n"
                "".format(energyK, energyA, energy, energy - self.energy_temp)
            )

            if self.output._has_to_save:
                memory = get_memory_usage()
                self._write_memory_txt()

                to_print += f"              memory  = {memory:9.3f} Mo.\n"

            duration_left = self._evaluate_duration_left()
            if duration_left is not None:
                to_print += f"              estimated remaining duration = {duration_left}"

        self.energy_temp = energy
        return to_print

    def _write_memory_txt(self):
        """Write memory .txt"""
        it = self.sim.time_stepping.it
        mem = get_memory_usage()
        self.file_memory.write(f"{it:.3f},{mem:.3f}\n")
        self.file_memory.flush()
        os.fsync(self.file_memory.fileno())

    def plot_memory(self):
        """ Plot memory used from memory_out.txt """
        with open(self.output.path_run + "/memory_out.txt") as file_memory:
            lines = file_memory.readlines()

        lines_it = []
        lines_memory = []
        for il, line in enumerate(lines):
            lines_it.append(float(line.split(",")[0]))
            lines_memory.append(float(line.split(",")[1]))

        fig, ax = self.output.figure_axe()
        ax.set_xlabel("it")
        ax.set_ylabel("Memory (Mo)")

        ax.plot(lines_it, lines_memory, "k", linewidth=2)
        return fig

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}

        with open(self.output.path_run + "/stdout.txt") as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        lines_EK = []
        lines_EA = []
        for il, line in enumerate(lines):
            if line.startswith("it ="):
                lines_t.append(line)
            if line.startswith("              energy  ="):
                lines_E.append(line)
            if line.startswith("              energyK ="):
                lines_EK.append(line)
            if line.startswith("              energyA ="):
                lines_EA.append(line)

        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        it = np.zeros(nt, dtype=int)
        t = np.zeros(nt)
        deltat = np.zeros(nt)

        E = np.zeros(nt)
        deltaE = np.zeros(nt)
        EK = np.zeros(nt)
        EA = np.zeros(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            it[il] = int(words[2])
            t[il] = float(words[6])
            deltat[il] = float(words[10])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])
            deltaE[il] = float(words[7])

            line = lines_EK[il]
            words = line.split()
            EK[il] = float(words[2])

            line = lines_EA[il]
            words = line.split()
            EA[il] = float(words[2])

        dict_results["it"] = it
        dict_results["t"] = t
        dict_results["deltat"] = deltat
        dict_results["E"] = E
        dict_results["deltaE"] = deltaE
        dict_results["EK"] = EK
        dict_results["EA"] = EA

        return dict_results

    def plot(self):
        dict_results = self.load()

        t = dict_results["t"]
        deltat = dict_results["deltat"]
        E = dict_results["E"]
        #  deltaE = dict_results['deltaE']
        EK = dict_results["EK"]
        EA = dict_results["EA"]

        x_left_axe = 0.12
        z_bottom_axe = 0.55
        width_axe = 0.85
        height_axe = 0.4
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("t")
        ax1.set_ylabel("deltat(t)")

        ax1.set_title("info stdout\n" + self.output.summary_simul)
        ax1.plot(t, deltat, "k", linewidth=2)

        size_axe[1] = 0.08
        ax2 = fig.add_axes(size_axe)
        ax2.set_xlabel("t")
        ax2.set_ylabel("E(t), deltaE(t)")
        ax2.plot(t, E, "k", linewidth=2, label="$E$")
        # ax2.plot(t, deltaE, 'b', linewidth=2)
        ax2.plot(t, EK, "r", linewidth=2, label="$E_K$")
        ax2.plot(t, EA, "g", linewidth=2, label="$E_A$")
        ax2.legend()
        ax2.grid(True)

    def close(self):
        super().close()
        try:
            self.file_memory.close()
        except AttributeError:
            pass



import numpy as np

from fluidsim.base.output.print_stdout import PrintStdOutBase

from fluiddyn.util import mpi


class PrintStdOutSW1L(PrintStdOutBase):
    """A :class:`PrintStdOutBase` object is used to print in both the
    stdout and the stdout.txt file, and also to print simple info on
    the current state of the simulation."""

    def _make_str_info(self):
        to_print = super()._make_str_info()

        energyK, energyA = self.output.compute_energiesKA()
        energy = energyK + energyA

        energy = self.output.compute_energy()
        if mpi.rank == 0:
            to_print += (
                "              energy  = {:9.3e} ; Delta energy = {:+9.3e}\n"
                "              energyK = {:9.3e} ; energyA      = {:9.3e}\n"
                "".format(energy, energy - self.energy_temp, energyK, energyA)
            )

            duration_left = self._evaluate_duration_left()
            if duration_left is not None:
                to_print += f"              estimated remaining duration = {duration_left}"

        self.energy_temp = energy
        return to_print

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}

        with open(self.output.path_run + "/stdout.txt") as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        lines_E_KA = []
        for il, line in enumerate(lines):
            if line[0:4] == "it =":
                lines_t.append(line)
            if line[0:23] == "              energy  =":
                lines_E.append(line)
            if line[0:23] == "              energyK =":
                lines_E_KA.append(line)
        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        it = np.zeros(nt, dtype=int)
        t = np.zeros(nt)
        deltat = np.zeros(nt)

        E = np.zeros(nt)
        deltaE = np.zeros(nt)

        E_K = np.zeros(nt)
        E_A = np.zeros(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            it[il] = int(words[2])
            t[il] = float(words[6])
            deltat[il] = float(words[10])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])
            deltaE[il] = float(words[7])

            line = lines_E_KA[il]
            words = line.split()
            E_K[il] = float(words[2])
            E_A[il] = float(words[6])

        dict_results["it"] = it
        dict_results["t"] = t
        dict_results["deltat"] = deltat
        dict_results["E"] = E
        dict_results["deltaE"] = deltaE
        dict_results["E_K"] = E_K
        dict_results["E_A"] = E_A
        return dict_results

    def plot(self):
        dict_results = self.load()

        t = dict_results["t"]
        deltat = dict_results["deltat"]
        E = dict_results["E"]
        deltaE = dict_results["deltaE"]
        E_K = dict_results["E_K"]
        E_A = dict_results["E_A"]

        x_left_axe = 0.12
        z_bottom_axe = 0.55
        width_axe = 0.85
        height_axe = 0.4
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("t")
        ax1.set_ylabel("deltat(t)")
        ax1.set_title("info stdout\n" + self.output.summary_simul)
        ax1.plot(t, deltat, "k", linewidth=2)

        z_bottom_axe = 0.08
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        ax2 = fig.add_axes(size_axe)
        ax2.set_xlabel("t")
        ax2.set_ylabel("E(t), deltaE(t)")
        ax2.plot(t, E, "k", linewidth=2)
        ax2.plot(t, E_K, "r", linewidth=2)
        ax2.plot(t, E_A, "b", linewidth=2)
        ax2.plot(t, deltaE, "k--", linewidth=2)



"""2D waves solver (:mod:`fluidsim.solvers.waves2d.solver`)
==================================================================

.. autoclass:: Simul
   :members:
   :private-members:

"""

import numpy as np
from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)


class InfoSolverWaves2d(InfoSolverPseudoSpectral):
    def _init_root(self):

        super()._init_root()

        package = "fluidsim.solvers.waves2d"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "Waves2d"

        classes = self.classes

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateWaves"


class Simul(SimulBasePseudoSpectral):
    r"""Pseudo-spectral solver for equations of 2D waves.

    Notes
    -----

    .. |p| mathmacro:: \partial

    This class is dedicated to solve wave 2D equations:

    .. math::
       \p_t \hat f = \hat g - \gamma_f \hat f,

       \p_t \hat g = -\Omega^2 \hat f - \gamma_g \hat g,

    This purely linear wave equation can alternatively be written as
    as :math:`\p_t X = M X`, with

    .. math::

       X = \begin{pmatrix} \hat f \\ \hat g \end{pmatrix}
       \ \ \text{and}\ \
       M = \begin{pmatrix} -\gamma_f & 1 \\
       -\Omega^2 & -\gamma_g \end{pmatrix},

    where the three coefficients usually depend on the wavenumber.
    The eigenvalues are :math:`\sigma_\pm = - \bar \gamma \pm i \tilde
    \Omega`, where :math:`\bar \gamma = (\gamma_f + \gamma_g)/2` and

    .. math::

       \tilde \Omega = \Omega \sqrt{1 +
       \frac{1}{\Omega^2}(\gamma_f\gamma_g - \bar \gamma^2)}.

    The (not normalized) eigenvectors can be expressed as

    .. math::

       V_\pm = \begin{pmatrix} 1 \\ \sigma_\pm + \gamma_f \end{pmatrix}.

    The state can be represented by a vector :math:`A` that verifies
    :math:`X = V A`, where :math:`V` is the base matrix

    .. math::

       V = \begin{pmatrix} 1 & 1 \\
       \sigma_+ + \gamma_f & \sigma_- + \gamma_f \end{pmatrix}.

    The inverse base matrix is given by

    .. math::

       V^{-1} = \frac{i}{2\tilde \Omega}
       \begin{pmatrix}
       \sigma_- + \gamma_f & -1 \\
       -\sigma_+ - \gamma_f &  1 \end{pmatrix}.


    """
    InfoSolver = InfoSolverWaves2d

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        SimulBasePseudoSpectral._complete_params_with_default(params)
        attribs = {"c2": 1.0, "f": 0}
        params._set_attribs(attribs)

    def tendencies_nonlin(self, state_spect=None, old=None):

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        tendencies_fft[:] = 0.0

        return tendencies_fft

    def compute_freq_complex(self, key):
        assert key in ["f_fft", "g_fft"], "Unexpected key: " + key
        if key == "f_fft":
            omega = self.oper.create_arrayK(value=0)
        elif key == "g_fft":
            omega = 1.0j * np.sqrt(
                self.params.f ** 2 + self.params.c2 * self.oper.K2
            )
        return omega


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 32
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    # params.oper.type_fft = 'FFTWPY'

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 1.0
    params.time_stepping.USE_CFL = False

    params.init_fields.type = "noise"

    # params.forcing.enable = True
    # params.forcing.type = 'tcrandom'
    # 'Proportional'
    # params.forcing.type_normalize

    # params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 0.5

    params.output.periods_plot.phys_fields = 0.0

    params.output.ONLINE_PLOT_OK = True

    # params.output.spectra.HAS_TO_PLOT_SAVED = True
    # params.output.spatial_means.HAS_TO_PLOT_SAVED = True
    # params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True
    # params.output.increments.HAS_TO_PLOT_SAVED = True

    params.output.phys_fields.field_to_plot = "f"

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    # sim.output.phys_fields.plot()

    fld.show()



"""State for the NS3D solver (:mod:`fluidsim.solvers.ns3d.state`)
=======================================================================
"""

import numpy as np


from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi


class StateNS3D(StatePseudoSpectral):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the solver NS3D.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        keys_state_phys = ["vx", "vy", "vz"]

        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": [k + "_fft" for k in keys_state_phys],
                "keys_state_phys": keys_state_phys,
                "keys_phys_needed": keys_state_phys,
                "keys_computable": ["rotz", "divh", "rotz_fft", "divh_fft"],
                "keys_linear_eigenmodes": ["rot_fft"],
            }
        )

    def __init__(self, sim, oper=None):

        super().__init__(sim, oper)

        self.fields_tmp = tuple(
            np.empty_like(self.state_phys[0]) for n in range(6)
        )

        self.fields_spect_tmp = tuple(
            np.empty_like(self.state_spect[0]) for n in range(3)
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "rotz":
            vx_fft = self.get_var("vx_fft")
            vy_fft = self.get_var("vy_fft")
            rotz_fft = self.oper.rotzfft_from_vxvyfft(vx_fft, vy_fft)
            result = self.oper.ifft3d(rotz_fft)
        elif key == "divh":
            vx_fft = self.get_var("vx_fft")
            vy_fft = self.get_var("vy_fft")
            divh_fft = self.oper.divhfft_from_vxvyfft(vx_fft, vy_fft)
            result = self.oper.ifft3d(divh_fft)
        elif key == "rotz_fft":
            vx_fft = self.get_var("vx_fft")
            vy_fft = self.get_var("vy_fft")
            result = self.oper.rotzfft_from_vxvyfft(vx_fft, vy_fft)
        elif key == "divh_fft":
            vx_fft = self.get_var("vx_fft")
            vy_fft = self.get_var("vy_fft")
            result = self.oper.divhfft_from_vxvyfft(vx_fft, vy_fft)
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                mpi.printby0(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def init_from_vxvyfft(self, vx_fft, vy_fft):
        self.state_spect.set_var("vx_fft", vx_fft)
        self.state_spect.set_var("vy_fft", vy_fft)
        self.state_spect.set_var("vz_fft", np.zeros_like(vx_fft))

        self.statephys_from_statespect()
        self.statespect_from_statephys()

    def init_from_vxvyvzfft(self, vx_fft, vy_fft, vz_fft):
        self.state_spect.set_var("vx_fft", vx_fft)
        self.state_spect.set_var("vy_fft", vy_fft)
        self.state_spect.set_var("vz_fft", vz_fft)

        self.statephys_from_statespect()
        self.statespect_from_statephys()

    def init_statespect_from(self, **kwargs):
        """Initialize `state_spect` from arrays.

        Parameters
        ----------

        **kwargs : {key: array, ...}

          keys and arrays used for the initialization. The other keys
          are set to zero.

        Examples
        --------

        .. code-block:: python

           kwargs = {'a_fft': Fa_fft}
           init_statespect_from(**kwargs)

           ux_fft, uy_fft, eta_fft = oper.uxuyetafft_from_qfft(q_fft)
           init_statespect_from(ux_fft=ux_fft, uy_fft=uy_fft, eta_fft=eta_fft)

        """

        if len(kwargs) == 1 and next(iter(kwargs.keys())) == "rotz_fft":
            self.state_spect[:] = 0.0
            vx_fft, vy_fft = self.oper.vxvyfft_from_rotzfft(kwargs["rotz_fft"])
            super().init_statespect_from(vx_fft=vx_fft, vy_fft=vy_fft)
        else:
            return super().init_statespect_from(**kwargs)



"""Plate2d solvers (:mod:`fluidsim.solvers.plate2d`)
==========================================================


Provides:

.. autosummary::
   :toctree:

   solver
   state
   operators
   output
   init_fields
   forcing
   dimensional

"""



"""Operators 2d (:mod:`fluidsim.operators.operators2d`)
=======================================================

Provides

.. autoclass:: OperatorsPseudoSpectral2D
   :members:
   :private-members:

"""

from warnings import warn
from random import uniform

import numpy as np

from transonic import boost, Array, Transonic
from fluiddyn.util import mpi
from fluidfft.fft2d.operators import OperatorsPseudoSpectral2D as _Operators

from fluidsim.base.params import Parameters
from ..base.setofvariables import SetOfVariables
from .. import _is_testing
from .base import OperatorBase

ts = Transonic()

if not ts.is_transpiling and not ts.is_compiled and not _is_testing:
    warn(
        "operators2d.py has to be pythranized to be efficient! "
        "Install pythran and recompile."
    )
elif ts.is_transpiling:
    _Operators = object


Af = Array[np.float64, "2d"]
Ac = Array[np.complex128, "2d"]


@boost
def laplacian_fft(a_fft: Ac, Kn: Af):
    """Compute the n-th order Laplacian."""
    return a_fft * Kn


@boost
def invlaplacian_fft(a_fft: Ac, Kn_not0: Af, rank: int):
    """Compute the n-th order inverse Laplacian."""
    invlap_afft = a_fft / Kn_not0
    if rank == 0:
        invlap_afft[0, 0] = 0.0
    return invlap_afft


@boost
def compute_increments_dim1(var: Af, irx: int):
    """Compute the increments of var over the dim 1."""
    n1 = var.shape[1]
    n1new = n1 - irx
    # bug for gast 0.4.0 (https://github.com/serge-sans-paille/gast/issues/48)
    inc_var = var[:, irx:n1] - var[:, 0:n1new]
    return inc_var


if not ts.is_transpiling:
    nb_proc = mpi.nb_proc
    rank = mpi.rank
else:
    nb_proc = 1
    rank = 0

if nb_proc > 1:
    MPI = mpi.MPI
    comm = mpi.comm


@boost
class OperatorsPseudoSpectral2D(_Operators, OperatorBase):

    _has_to_dealiase: bool
    where_dealiased: "uint8[:, :]"
    KX: Af
    KY: Af
    deltax: float
    deltay: float

    @classmethod
    def _create_default_params(cls):
        params = Parameters(tag="params", attribs={"ONLY_COARSE_OPER": False})
        cls._complete_params_with_default(params)
        return params

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        attribs = {
            "type_fft": "default",
            "coef_dealiasing": 2.0 / 3,
            "nx": 48,
            "ny": 48,
            "Lx": 8,
            "Ly": 8,
            "truncation_shape": "cubic",
            "NO_SHEAR_MODES": False,
            "NO_KY0": False,
        }
        params._set_child("oper", attribs=attribs)

    def __init__(self, params):

        self.params = params
        self.axes = ("y", "x")
        nx = int(params.oper.nx)
        ny = int(params.oper.ny)

        if params.oper.nx != nx:
            raise ValueError(
                "params.oper.nx != int(params.oper.nx); "
                "({})".format(params.oper.nx)
            )

        if params.oper.ny != ny:
            raise ValueError(
                "params.oper.ny != int(params.oper.ny); "
                "({})".format(params.oper.ny)
            )

        params.oper.nx = nx
        params.oper.ny = ny

        if params.ONLY_COARSE_OPER:
            nx = ny = 4

        super().__init__(
            nx,
            ny,
            params.oper.Lx,
            params.oper.Ly,
            fft=params.oper.type_fft,
            coef_dealiasing=params.oper.coef_dealiasing,
        )

        # compatibility for fluidfft <= 0.3.0
        if not hasattr(self, "oper_fft") and hasattr(self, "opfft"):
            self.oper_fft = self.opfft

        self.Lx = self.lx
        self.Ly = self.ly

        try:
            self.project_fft_on_realX = self._opfft.project_fft_on_realX
        except AttributeError:
            if self.is_sequential:
                self.project_fft_on_realX = self.project_fft_on_realX_seq
            else:
                self.project_fft_on_realX = self.project_fft_on_realX_slow

        if not self.is_sequential:
            self.iKxloc_start, _ = self.oper_fft.get_seq_indices_first_K()
            self.iKxloc_start_rank = np.array(comm.allgather(self.iKxloc_start))

            nkx_loc_rank = np.array(comm.allgather(self.nkx_loc))
            a = nkx_loc_rank
            self.SAME_SIZE_IN_ALL_PROC = (a >= a.max()).all()

        self._reinit_truncation()

        try:
            NO_SHEAR_MODES = self.params.oper.NO_SHEAR_MODES
        except AttributeError:
            pass
        else:
            if NO_SHEAR_MODES:
                COND_NOSHEAR = abs(self.KX) == 0.0
                self.where_dealiased = np.array(
                    np.logical_or(COND_NOSHEAR, self.where_dealiased),
                    dtype=np.uint8,
                )

        try:
            NO_KY0 = self.params.oper.NO_KY0
        except AttributeError:
            pass
        else:
            if NO_KY0:
                COND_NO_KY0 = abs(self.KY) == 0.0
                self.where_dealiased = np.array(
                    np.logical_or(COND_NO_KY0, self.where_dealiased),
                    dtype=np.uint8,
                )

    def get_region_multiple_aliases(self):
        aliases_x = abs(self.KX) >= 2 / 3 * self.deltakx * self.nx / 2
        aliases_y = abs(self.KY) >= 2 / 3 * self.deltaky * self.ny / 2
        return aliases_x & aliases_y

    def dealiasing(self, *args):
        if not self._has_to_dealiase:
            return

        for thing in args:
            if isinstance(thing, SetOfVariables):
                self.dealiasing_setofvar(thing)
            elif isinstance(thing, np.ndarray):
                self.dealiasing_variable(thing)

    @boost
    def dealiasing_setofvar(self, sov: "complex128[][][]"):
        """Dealiasing of a setofvar arrays."""
        if self._has_to_dealiase:
            nk, n0, n1 = sov.shape

            for i0 in range(n0):
                for i1 in range(n1):
                    if self.where_dealiased[i0, i1]:
                        for ik in range(nk):
                            sov[ik, i0, i1] = 0.0

    def project_fft_on_realX_seq(self, f_fft):
        """Project the given field in spectral space such as its
        inverse fft is a real field."""

        nky_seq = self.shapeK_seq[0]

        iky_ky0 = 0
        iky_kyM = nky_seq // 2
        ikx_kx0 = 0
        # ikx_kxM = self.nkx_seq-1
        ikx_kxM = self.shapeK_seq[1] - 1

        # first, some values have to be real
        f_fft[iky_ky0, ikx_kx0] = f_fft[iky_ky0, ikx_kx0].real
        f_fft[iky_ky0, ikx_kxM] = f_fft[iky_ky0, ikx_kxM].real
        f_fft[iky_kyM, ikx_kx0] = f_fft[iky_kyM, ikx_kx0].real
        f_fft[iky_kyM, ikx_kxM] = f_fft[iky_kyM, ikx_kxM].real

        # second, there are relations between some values
        for ikyp in range(1, iky_kyM):
            ikyn = nky_seq - ikyp

            f_kp_kx0 = f_fft[ikyp, ikx_kx0]
            f_kn_kx0 = f_fft[ikyn, ikx_kx0]

            f_fft[ikyp, ikx_kx0] = (f_kp_kx0 + f_kn_kx0.conjugate()) / 2
            f_fft[ikyn, ikx_kx0] = (
                (f_kp_kx0 + f_kn_kx0.conjugate()) / 2
            ).conjugate()

            f_kp_kxM = f_fft[ikyp, ikx_kxM]
            f_kn_kxM = f_fft[ikyn, ikx_kxM]

            f_fft[ikyp, ikx_kxM] = (f_kp_kxM + f_kn_kxM.conjugate()) / 2
            f_fft[ikyn, ikx_kxM] = (
                (f_kp_kxM + f_kn_kxM.conjugate()) / 2
            ).conjugate()
        return f_fft

    def project_fft_on_realX_slow(self, f_fft):
        return self.fft(self.ifft(f_fft))

    def coarse_seq_from_fft_loc(self, f_fft, shapeK_loc_coarse):
        """Return a coarse field in K space."""
        nkyc = shapeK_loc_coarse[0]
        nkxc = shapeK_loc_coarse[1]

        if nb_proc > 1:
            if not self.is_transposed:
                raise NotImplementedError()

            fc_trans = np.empty([nkxc, nkyc], np.complex128)
            nky = self.shapeK_seq[1]
            f1d_temp = np.empty([nkyc], np.complex128)

            for ikxc in range(nkxc):
                kx = self.deltakx * ikxc
                rank_ikx, ikxloc, ikyloc = self.where_is_wavenumber(kx, 0.0)
                if rank == rank_ikx:
                    # create f1d_temp
                    for ikyc in range(nkyc):
                        if ikyc <= nkyc / 2:
                            iky = ikyc
                        else:
                            kynodim = ikyc - nkyc
                            iky = kynodim + nky
                        f1d_temp[ikyc] = f_fft[ikxloc, iky]

                if rank_ikx != 0:
                    # message f1d_temp
                    if rank == 0:
                        # print('f1d_temp', f1d_temp, f1d_temp.dtype)
                        comm.Recv(
                            [f1d_temp, MPI.DOUBLE_COMPLEX],
                            source=rank_ikx,
                            tag=ikxc,
                        )
                    elif rank == rank_ikx:
                        comm.Send(
                            [f1d_temp, MPI.DOUBLE_COMPLEX], dest=0, tag=ikxc
                        )
                if rank == 0:
                    # copy into fc_trans
                    fc_trans[ikxc] = f1d_temp.copy()
            fc_fft = fc_trans.transpose()

        else:
            nky = self.shapeK_seq[0]
            fc_fft = np.empty([nkyc, nkxc], np.complex128)
            for ikyc in range(nkyc):
                if ikyc <= nkyc / 2:
                    iky = ikyc
                else:
                    kynodim = ikyc - nkyc
                    iky = kynodim + nky
                for ikxc in range(nkxc):
                    fc_fft[ikyc, ikxc] = f_fft[iky, ikxc]

        # fc_fft[nkyc//2] *= 2

        # energy_coarse = self.sum_wavenumbers(abs(fc_fft)**2)
        # energy_global = self.sum_wavenumbers(abs(f_fft)**2)
        # print('energy_coarse = {}'.format(energy_coarse))
        # print('energy_global = {}'.format(energy_global))

        # assert energy_global == energy_coarse

        return fc_fft

    # def fft_loc_from_coarse_seq(self, fc_fft, shapeK_loc_coarse):
    #     """Return a large field in K space."""
    #     nkyc = shapeK_loc_coarse[0]
    #     nkxc = shapeK_loc_coarse[1]

    #     if nb_proc > 1:
    #         nky = self.shapeK_seq[0]
    #         f_fft = self.create_arrayK(value=0.)
    #         fc_trans = fc_fft.transpose()

    #         for ikxc in range(nkxc):
    #             kx = self.deltakx*ikxc
    #             rank_ikx, ikxloc, ikyloc = self.where_is_wavenumber(kx, 0.)
    #             fc1D = fc_trans[ikxc]
    #             if rank_ikx != 0:
    #                 # message fc1D
    #                 fc1D = np.ascontiguousarray(fc1D)
    #                 if rank == 0:
    #                     comm.Send(fc1D, dest=rank_ikx, tag=ikxc)
    #                 elif rank == rank_ikx:
    #                     comm.Recv(fc1D, source=0, tag=ikxc)
    #             if rank == rank_ikx:
    #                 # copy
    #                 for ikyc in range(nkyc):
    #                     if ikyc <= nkyc/2:
    #                         iky = ikyc
    #                     else:
    #                         kynodim = ikyc - nkyc
    #                         iky = kynodim + nky
    #                     f_fft[ikxloc, iky] = fc1D[ikyc]

    #     else:
    #         nky = self.shapeK_seq[0]
    #         nkx = self.shapeK_seq[1]
    #         f_fft = np.zeros([nky, nkx], np.complex128)
    #         for ikyc in range(nkyc):
    #             if ikyc <= nkyc/2:
    #                 iky = ikyc
    #             else:
    #                 kynodim = ikyc - nkyc
    #                 iky = kynodim + nky
    #             for ikxc in range(nkxc):
    #                 f_fft[iky, ikxc] = fc_fft[ikyc, ikxc]
    #     return f_fft

    def compute_increments_dim1(self, var, irx):
        """Compute the increments of var over the dim 1."""
        return compute_increments_dim1(var, int(irx))

    def pdf_normalized(self, field, nb_bins=100):
        """Compute the normalized pdf"""

        field_max = field.max()
        field_min = field.min()

        if nb_proc > 1:
            field_max = comm.allreduce(field_max, op=MPI.MAX)
            field_min = comm.allreduce(field_min, op=MPI.MIN)

        range_min = field_min
        range_max = field_max

        if nb_proc == 1:
            pdf, bin_edges = np.histogram(
                field, bins=nb_bins, density=True, range=(range_min, range_max)
            )
        else:
            hist, bin_edges = np.histogram(
                field, bins=nb_bins, range=(range_min, range_max)
            )
            # memory leak related to this line for CPython 3.7.1
            # hist = comm.allreduce(hist, op=MPI.SUM)
            # workaround for CPython 3.7.0 and 3.7.1
            tmp = np.empty_like(hist)
            comm.Allreduce(hist, tmp, op=MPI.SUM)
            hist = tmp
            #
            pdf = hist / ((bin_edges[1] - bin_edges[0]) * hist.sum())
        return pdf, bin_edges

    def where_is_wavenumber(self, kx_approx, ky_approx):
        ikx_seq = int(np.round(kx_approx / self.deltakx))
        if ikx_seq >= self.nkx_seq:
            raise ValueError("not good :-) ikx_seq >= self.nkx_seq")

        if self.is_sequential:
            rank_k = 0
            ikx_loc = ikx_seq
        else:
            if self.SAME_SIZE_IN_ALL_PROC:
                rank_k = int(np.floor(float(ikx_seq) / self.nkx_loc))
                if ikx_seq >= self.nkx_loc * mpi.nb_proc:
                    raise ValueError(
                        "not good :-) ikx_seq >= self.nkx_loc * mpi.nb_proc\n"
                        "ikx_seq, self.nkx_loc, mpi.nb_proc = "
                        "{}, {}, {}".format(ikx_seq, self.nkx_loc, mpi.nb_proc)
                    )

            else:
                rank_k = 0
                while rank_k < self.nb_proc - 1 and (
                    not (
                        self.iKxloc_start_rank[rank_k] <= ikx_seq
                        and ikx_seq < self.iKxloc_start_rank[rank_k + 1]
                    )
                ):
                    rank_k += 1

            ikx_loc = ikx_seq - self.iKxloc_start_rank[rank_k]

        iky_loc = int(np.round(ky_approx / self.deltaky))
        if iky_loc < 0:
            iky_loc = self.nky_loc + iky_loc

        if self.is_transposed:
            ik0_loc = ikx_loc
            ik1_loc = iky_loc
        else:
            ik0_loc = iky_loc
            ik1_loc = ikx_loc

        return rank_k, ik0_loc, ik1_loc

    def uxuyfft_from_psifft(self, psi_fft):
        px_psi_fft, py_psi_fft = self.gradfft_from_fft(psi_fft)
        ux_fft = -py_psi_fft
        uy_fft = px_psi_fft
        return ux_fft, uy_fft

    def rotfft_from_psifft(self, psi_fft):
        return self.laplacian_fft(psi_fft)

    def pxffft_from_fft(self, f_fft):
        """Return the gradient of f_fft in spectral space."""
        return 1j * self.KX * f_fft

    def pyffft_from_fft(self, f_fft):
        """Return the gradient of f_fft in spectral space."""
        return 1j * self.KY * f_fft

    def laplacian2_fft(self, a_fft):
        warn("Use oper.laplacian_fft instead.", PendingDeprecationWarning)
        return laplacian_fft(a_fft, self.K4)

    def invlaplacian2_fft(self, a_fft):
        warn("Use oper.invlaplacian_fft instead.", PendingDeprecationWarning)
        return invlaplacian_fft(a_fft, self.K4_not0, rank)

    def laplacian_fft(self, a_fft, order=2, negative=False):
        r"""Compute the n-th order Laplacian, :math:`\nabla^{n} \hat{a}`

        Parameters
        ----------
        a_fft : ndarray

        order: int, {2, 4, 8}, optional
            Order of the Laplacian operator

        negative: bool, optional
            Negative of the result.

        """
        sign = 1j ** order
        if sign.imag != 0:
            raise ValueError(f"Order={order} should be even!")

        if negative:
            sign *= -1
        Kn = getattr(self, "K{}".format(int(order)))

        # Avoid unnecessary multiplication by unity
        if sign == 1:
            return laplacian_fft(a_fft, Kn)
        else:
            return sign * laplacian_fft(a_fft, Kn)

    def invlaplacian_fft(self, a_fft, order=2, negative=False):
        r"""Compute the n-th order inverse Laplacian, :math:`\nabla^{-n} \hat{a}`

        Parameters
        ----------
        a_fft : ndarray

        order: int, {2, 4, 8}, optional
            Order of the inverse Laplacian operator.

        negative: bool, optional
            Negative of the result.

        """
        sign = 1.0 / 1j ** order
        if sign.imag != 0:
            raise ValueError(f"Order={order} should be even!")

        if negative:
            sign *= -1
        Kn_not0 = getattr(self, "K{}_not0".format(int(order)))

        # Avoid unnecessary multiplication by unity
        if sign == 1:
            return invlaplacian_fft(a_fft, Kn_not0, rank)
        else:
            return sign * invlaplacian_fft(a_fft, Kn_not0, rank)

    def put_coarse_array_in_array_fft(
        self, arr_coarse, arr, oper_coarse, shapeK_loc_coarse
    ):
        """Put the values contained in a coarse array in an array.

        Both arrays are in Fourier space.

        """
        if arr.ndim == 3:
            if mpi.rank == 0:
                if arr_coarse.ndim != 3:
                    raise ValueError

            for ikey in range(arr.shape[0]):
                if mpi.rank == 0:
                    arr2d_coarse = arr_coarse[ikey]
                else:
                    arr2d_coarse = None
                self.put_coarse_array_in_array_fft(
                    arr2d_coarse, arr[ikey], oper_coarse, shapeK_loc_coarse
                )
            return

        nKyc, nKxc = shapeK_loc_coarse

        if mpi.nb_proc > 1 and not self.is_sequential:
            if not self.is_transposed:
                raise NotImplementedError()

            nKy = self.shapeK_seq[1]

            if mpi.rank == 0:
                fck_fft = arr_coarse.transpose()

            for ikxc in range(nKxc):
                kx = self.deltakx * ikxc
                rank_ikx, ikxloc, ikyloc = self.where_is_wavenumber(kx, 0.0)

                if mpi.rank == 0:
                    fc1D = fck_fft[ikxc]

                if rank_ikx != 0:
                    # message fc1D
                    if mpi.rank == rank_ikx:
                        fc1D = np.empty([nKyc], dtype=np.complex128)
                    if mpi.rank == 0 or mpi.rank == rank_ikx:
                        fc1D = np.ascontiguousarray(fc1D)
                    if mpi.rank == 0:
                        mpi.comm.Send(
                            [fc1D, mpi.MPI.COMPLEX], dest=rank_ikx, tag=ikxc
                        )
                    elif mpi.rank == rank_ikx:
                        mpi.comm.Recv([fc1D, mpi.MPI.COMPLEX], source=0, tag=ikxc)
                if mpi.rank == rank_ikx:
                    # copy
                    for ikyc in range(nKyc):
                        if ikyc <= nKyc / 2.0:
                            iky = ikyc
                        else:
                            kynodim = ikyc - nKyc
                            iky = kynodim + nKy
                        arr[ikxloc, iky] = fc1D[ikyc]
        else:
            nKy = self.shapeK_seq[0]

            if not np.allclose(0.0, max(abs(arr_coarse[nKyc // 2, :]))):
                raise ValueError("any(arr_coarse[nKyc//2] != 0)")

            if not np.allclose(0.0, max(abs(arr_coarse[:, nKxc - 1]))):
                raise ValueError("any(arr_coarse[:, nKxc-1] != 0)")

            for ikyc in range(nKyc):
                if ikyc <= nKyc / 2.0:
                    iky = ikyc
                else:
                    kynodim = ikyc - nKyc
                    iky = kynodim + nKy

                for ikxc in range(nKxc):
                    arr[iky, ikxc] = arr_coarse[ikyc, ikxc]

    def get_grid1d_seq(self, axe="x"):

        if axe not in ("x", "y"):
            raise ValueError

        if self.params.ONLY_COARSE_OPER:
            number_points = getattr(self.params.oper, "n" + axe)
            length = getattr(self, "L" + axe)
            return np.linspace(0, length, number_points)
        else:
            return getattr(self, axe + "_seq")

    @boost
    def get_phases_random(self):
        # Not supported by Pythran 0.9.5!
        # alpha_x, alpha_y = np.random.uniform(-0.5, 0.5, 2)
        alpha_x, alpha_y = tuple(uniform(-0.5, 0.5) for _ in range(2))
        beta_x = alpha_x + 0.5 if alpha_x < 0 else alpha_x - 0.5
        beta_y = alpha_y + 0.5 if alpha_y < 0 else alpha_y - 0.5

        phase_alpha = (
            alpha_x * self.deltax * self.KX + alpha_y * self.deltay * self.KY
        )
        phase_beta = (
            beta_x * self.deltax * self.KX + beta_y * self.deltay * self.KY
        )
        return phase_alpha, phase_beta


# energy_arr = self.sum_wavenumbers(abs(arr)**2)
# energy_array_coarse_after = oper_coarse.sum_wavenumbers(
#     abs(arr_coarse)**2)
# print('energy_array_coarse_after  = ', energy_array_coarse_after)
# print('energy_arr                 = ', energy_arr)



# !/usr/bin/env python
# coding=utf8
#
# run simul_test_conserveE.py
# mpirun -np 8 python simul_test_conserveE.py

key_solver = "NS2D"
# key_solver = 'SW1l'
# key_solver = 'SW1l.onlywaves'
# key_solver = 'SW1l.exactlin'
# key_solver = 'SW1l.modified'

import numpy as np
import fluidsim

solver = fluidsim.import_module_solver_from_key(key_solver)
params = solver.Simul.create_default_params()

params.short_name_type_run = "conservE"

nh = 16 * 4 * 8
Lh = 2 * np.pi
params.oper.nx = nh
params.oper.ny = nh
params.oper.Lx = Lh
params.oper.Ly = Lh

params.oper.coef_dealiasing = 2.0 / 3

params.nu_8 = 0.0
params.nu_4 = 0.0
params.nu_2 = 0.0
params.nu_m4 = 0.0

try:
    params.f = 1.0
    params.c2 = 200.0
except KeyError:
    pass

params.time_stepping.USE_CFL = False

params.time_stepping.it_end = 5
params.time_stepping.USE_T_END = False

params.time_stepping.type_time_scheme = "RK4"

params.init_fields.type_flow_init = "DIPOLE"


params.output.periods_print.print_stdout = 10.0e-15

params.output.periods_save.phys_fields = 0.0
params.output.periods_save.spectra = 0.0
params.output.periods_save.spect_energy_budg = 0.0
params.output.periods_save.increments = 0.0

try:
    params.output.periods_save.pdf = 0.0
    params.output.periods_save.time_signals_fft = False
except KeyError:
    pass


params.output.periods_plot.phys_fields = 0.0


params.time_stepping.deltat0 = 1.0e-1
sim = solver.Simul(params)
sim.time_stepping.start()

params.time_stepping.deltat0 = 1.0e-2
sim = solver.Simul(params)
sim.time_stepping.start()



"""NS2D solver (:mod:`fluidsim.solvers.ns2d.solver`)
====================================================

This module provides two classes defining the pseudo-spectral solver
2D incompressible Navier-Stokes equations (ns2d).

.. autoclass:: InfoSolverNS2D
   :members:
   :private-members:

.. autoclass:: Simul
   :members:
   :private-members:

"""

import sys

import numpy as np

from transonic import boost, Array

from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)

Af = Array[np.float64, "2d"]


@boost
def compute_Frot(ux: Af, uy: Af, px_rot: Af, py_rot: Af, beta: float = 0):
    if beta == 0:
        return -ux * px_rot - uy * py_rot

    else:
        return -ux * px_rot - uy * (py_rot + beta)


class InfoSolverNS2D(InfoSolverPseudoSpectral):
    """Contain the information on the solver ns2d.

    .. inheritance-diagram:: InfoSolverNS2D

    """

    def _init_root(self):
        """Init. `self` by writing the information on the solver.

        The function `InfoSolverPseudoSpectral._init_root` is
        called. We keep two classes listed by this function:

        - :class:`fluidsim.base.time_stepping.pseudo_spect.TimeSteppingPseudoSpectral`

        - :class:`fluidsim.operators.operators2d.OperatorsPseudoSpectral2D`


        The other first-level classes for this solver are:

        - :class:`fluidsim.solvers.ns2d.solver.Simul`

        - :class:`fluidsim.solvers.ns2d.state.StateNS2D`

        - :class:`fluidsim.solvers.ns2d.init_fields.InitFieldsNS2D`

        - :class:`fluidsim.solvers.ns2d.output.Output`

        - :class:`fluidsim.solvers.ns2d.forcing.ForcingNS2D`

        """
        super()._init_root()

        package = "fluidsim.solvers.ns2d"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "NS2D"

        classes = self.classes

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateNS2D"

        classes.InitFields.module_name = package + ".init_fields"
        classes.InitFields.class_name = "InitFieldsNS2D"

        classes.Output.module_name = package + ".output"
        classes.Output.class_name = "Output"

        classes.Forcing.module_name = package + ".forcing"
        classes.Forcing.class_name = "ForcingNS2D"


class Simul(SimulBasePseudoSpectral):
    """Pseudo-spectral solver 2D incompressible Navier-Stokes equations.

    .. inheritance-diagram:: Simul

    """

    InfoSolver = InfoSolverNS2D

    @staticmethod
    def _complete_params_with_default(params):
        """Complete the `params` container (static method)."""
        SimulBasePseudoSpectral._complete_params_with_default(params)
        attribs = {"beta": 0.0}
        params._set_attribs(attribs)

    def tendencies_nonlin(self, state_spect=None, old=None):
        r"""Compute the nonlinear tendencies.

        Parameters
        ----------

        state_spect : :class:`fluidsim.base.setofvariables.SetOfVariables`
            optional

            Array containing the state, i.e. the vorticity, in Fourier
            space.  If `state_spect`, the variables vorticity and the
            velocity are computed from it, otherwise, they are taken
            from the global state of the simulation, `self.state`.

            These two possibilities are used during the Runge-Kutta
            time-stepping.

        Returns
        -------

        tendencies_fft : :class:`fluidsim.base.setofvariables.SetOfVariables`
            An array containing the tendencies for the vorticity.

        Notes
        -----

        .. |p| mathmacro:: \partial

        The 2D Navier-Stokes equation can be written as:

        .. math:: \p_t \hat\zeta = \hat N(\zeta) - \sigma(k) \hat \zeta,

        This function compute the nonlinear term ("tendencies")
        :math:`N(\zeta) = - \mathbf{u}\cdot \mathbf{\nabla} \zeta`.

        """
        # the operator and the fast Fourier transform
        oper = self.oper
        ifft_as_arg = oper.ifft_as_arg

        # get or compute rot_fft, ux and uy
        if state_spect is None:
            rot_fft = self.state.state_spect.get_var("rot_fft")
            ux = self.state.state_phys.get_var("ux")
            uy = self.state.state_phys.get_var("uy")
        else:
            rot_fft = state_spect.get_var("rot_fft")
            ux_fft, uy_fft = oper.vecfft_from_rotfft(rot_fft)
            ux = self.state.field_tmp0
            uy = self.state.field_tmp1
            ifft_as_arg(ux_fft, ux)
            ifft_as_arg(uy_fft, uy)

        # "px" like $\partial_x$
        px_rot_fft, py_rot_fft = oper.gradfft_from_fft(rot_fft)

        px_rot = self.state.field_tmp2
        py_rot = self.state.field_tmp3

        ifft_as_arg(px_rot_fft, px_rot)
        ifft_as_arg(py_rot_fft, py_rot)

        Frot = compute_Frot(ux, uy, px_rot, py_rot, self.params.beta)

        if old is None:
            tendencies_fft = SetOfVariables(like=self.state.state_spect)
        else:
            tendencies_fft = old

        Frot_fft = tendencies_fft.get_var("rot_fft")
        oper.fft_as_arg(Frot, Frot_fft)

        oper.dealiasing(Frot_fft)

        # import numpy as np
        # T_rot = np.real(Frot_fft.conj()*rot_fft + Frot_fft*rot_fft.conj())/2.
        # print(('sum(T_rot) = {0:9.4e} ; sum(abs(T_rot)) = {1:9.4e}'
        #       ).format(self.oper.sum_wavenumbers(T_rot),
        #                self.oper.sum_wavenumbers(abs(T_rot))))

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft


if "sphinx" in sys.modules:
    params = Simul.create_default_params()

    __doc__ += (
        "Default parameters\n"
        "------------------\n"
        ".. code-block:: xml\n\n    "
        + "\n    ".join(params.__str__().split("\n\n", 1)[1].split("\n"))
        + "\n"
        + params._get_formatted_docs()
    )


if __name__ == "__main__":

    from math import pi

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    params.oper.nx = params.oper.ny = nh = 32
    params.oper.Lx = params.oper.Ly = Lh = 2 * pi
    # params.oper.coef_dealiasing = 1.

    delta_x = Lh / nh

    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 10.0

    params.init_fields.type = "dipole"

    params.forcing.enable = False
    params.forcing.type = "random"
    # 'Proportional'
    # params.forcing.type_normalize

    params.output.sub_directory = "tests"

    # params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 1.0
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spatial_means = 0.05
    # params.output.periods_save.spect_energy_budg = 0.5
    # params.output.periods_save.increments = 0.5

    # params.output.periods_plot.phys_fields = 2.0

    params.output.ONLINE_PLOT_OK = True

    # params.output.spectra.HAS_TO_PLOT_SAVED = True
    # params.output.spatial_means.HAS_TO_PLOT_SAVED = True
    # params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True
    # params.output.increments.HAS_TO_PLOT_SAVED = True

    params.output.phys_fields.field_to_plot = "rot"

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    # sim.output.phys_fields.plot()

    fld.show()



import unittest
import sys
from pathlib import Path
from math import pi

import pytest

import numpy as np
import matplotlib.pyplot as plt

import fluidsim as fls

import fluiddyn.util.mpi as mpi


from fluidsim import (
    modif_resolution_from_dir,
    load_state_phys_file,
    load_for_restart,
)
from fluidsim.base.output import run


from fluidsim.util.testing import TestSimul, skip_if_no_fluidfft, classproperty


@skip_if_no_fluidfft
class TestSimulBase(TestSimul):
    @classproperty
    def Simul(cls):
        from fluidsim.solvers.ns3d.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):

        params = cls.params = cls.Simul.create_default_params()

        params.short_name_type_run = "test"
        params.output.sub_directory = "unittests"
        nx = 32
        params.oper.nx = nx
        params.oper.ny = nx * 3 // 4

        Lx = 6.0
        params.oper.Lx = Lx
        params.oper.Ly = Lx * params.oper.ny / params.oper.nx
        try:
            params.oper.nz = nx // 2
            params.oper.Lz = Lx * params.oper.nz / params.oper.nx
        except AttributeError:
            pass

        params.oper.coef_dealiasing = 2.0 / 3
        params.nu_4 = 2.0
        params.nu_8 = 2.0

        params.time_stepping.t_end = 0.2
        params.init_fields.type = "noise"

        return params


class TestTendency(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.output.HAS_TO_SAVE = False

    def test_tendency(self):

        sim = self.sim
        tend = sim.tendencies_nonlin(state_spect=sim.state.state_spect)

        T_tot = np.ascontiguousarray(sim.oper.create_arrayK(value=0.0).real)

        for axis in ("x", "y", "z"):
            key = f"v{axis}_fft"
            vi_fft = sim.state.get_var(key)
            Fi_fft = tend.get_var(key)
            T_tot += np.real(Fi_fft.conj() * vi_fft)

        ratio = sim.oper.sum_wavenumbers(T_tot) / sim.oper.sum_wavenumbers(
            abs(T_tot)
        )

        self.assertGreater(1e-15, abs(ratio))


class TestOutput(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.oper.truncation_shape = "no_multiple_aliases"

        params.init_fields.type = "dipole"

        params.time_stepping.max_elapsed = 600
        params.time_stepping.type_time_scheme = "RK2_phaseshift_random"

        params.forcing.enable = True
        params.forcing.type = "in_script"
        params.forcing.key_forced = "vx_fft"

        params.nu_m4 = 1e-3

        # save all outputs!
        periods = params.output.periods_save
        for key in periods._key_attribs:
            periods[key] = 0.1

        Lx, Ly, Lz = params.oper.Lx, params.oper.Ly, params.oper.Lz
        nx, ny, nz = params.oper.nx, params.oper.ny, params.oper.nz
        probes_region = (0.0, Lx, 0.0, Ly, 0.0, Lz)
        params.output.temporal_spectra.probes_region = probes_region
        params.output.temporal_spectra.probes_deltax = Lx / nx
        params.output.temporal_spectra.probes_deltay = Ly / ny
        params.output.temporal_spectra.probes_deltaz = Lz / nz
        params.output.temporal_spectra.SAVE_AS_FLOAT32 = False

        nx, ny, nz = params.oper.nx, params.oper.ny, params.oper.nz
        params.output.spatiotemporal_spectra.probes_region = (
            nx // 2,
            ny // 2,
            nz // 2,
        )
        params.output.spatiotemporal_spectra.SAVE_AS_COMPLEX64 = False

    @pytest.mark.filterwarnings("ignore:divide by zero encountered in log10")
    def test_output(self):

        sim = self.sim

        # put energy in vz
        vz = sim.state.state_phys.get_var("vz")
        X, Y, Z = sim.oper.get_XYZ_loc()
        vz += 0.05 * np.cos(2 * pi * X / sim.oper.Lx)
        sim.state.statespect_from_statephys()

        sim.time_stepping.start()

        # testing phaseshift
        phaseshift = sim.time_stepping._get_phaseshift()
        assert phaseshift.shape == sim.oper.Kx.shape
        assert sim.time_stepping._get_phaseshift() is phaseshift

        if mpi.nb_proc == 1:

            phys_fields = sim.output.phys_fields
            phys_fields.plot(equation=f"iz=0", numfig=1000)

            phys_fields.get_field_to_plot_from_state()
            phys_fields.get_field_to_plot_from_state(sim.state.get_var("vx"))
            phys_fields.get_field_to_plot_from_state("vx", "x=0")
            phys_fields.get_field_to_plot_from_state("vx", "ix=0")
            phys_fields.get_field_to_plot_from_state("vx", "y=0")
            phys_fields.get_field_to_plot_from_state("vx", "iy=0")
            phys_fields.get_field_to_plot_from_state("vx", "z=0")

        # compute twice for better coverage
        sim.state.compute("rotz")
        sim.state.compute("rotz")
        sim.state.compute("divh")

        sim.output.phys_fields._get_grid1d("iz=0")
        sim.output.phys_fields._get_grid1d("iy=0")

        path_run = sim.output.path_run
        if mpi.nb_proc > 1:
            path_run = mpi.comm.bcast(path_run)

        if mpi.nb_proc == 1:
            sim2 = fls.load_sim_for_plot(path_run)
            sim2.output.plot_summary(tmin=0, key_field="vx")
            sim2.output.print_stdout.load()
            sim2.output.print_stdout.plot()
            sim2.output.spatial_means.load()
            sim2.output.spatial_means.load_dataset()
            sim2.output.spatial_means.plot(plot_injection=True, plot_hyper=True)
            sim2.output.spatial_means.plot_dt_E()
            sim2.output.spectra.load1d_mean()
            sim2.output.spectra.load3d_mean()
            sim2.output.spectra.plot1d(
                coef_plot_k2=1.0,
                coef_plot_k3=1.0,
                coef_plot_k53=1.0,
                xlim=(0.1, 1),
                ylim=(0.1, 1),
            )
            sim2.output.spectra.plot1d_times(
                tmin=0.1,
                tmax=10,
                delta_t=0.01,
                coef_compensate=5 / 3,
                coef_plot_k2=1.0,
                coef_plot_k3=1.0,
                coef_plot_k53=1.0,
                xlim=(0.1, 1),
                ylim=(0.1, 1),
            )
            sim2.output.spectra.plot3d_times(
                tmin=0.1,
                tmax=10,
                coef_compensate=5 / 3,
                coef_plot_k2=1.0,
                coef_plot_k3=1.0,
                coef_plot_k53=1.0,
            )
            sim2.output.spectra.plot3d_cumul_diss(tmin=0.1, tmax=10)

            sim2.output.phys_fields.set_equation_crosssection(
                f"x={sim.oper.Lx/4}"
            )
            sim2.output.phys_fields.animate("vx")

            sim2.output.phys_fields.plot(
                field="vx", time=10, equation=f"z={sim.oper.Lz/4}"
            )
            sim2.plot_freq_diss("z")

        sim3 = fls.load_state_phys_file(path_run, modif_save_params=False)
        sim3.params.time_stepping.t_end += 0.2
        sim3.time_stepping.start()

        if mpi.nb_proc > 1:
            plt.close("all")
            return

        sim3.output.phys_fields.animate(
            "vx",
            dt_frame_in_sec=1e-6,
            dt_equations=0.3,
            repeat=False,
            clim=(-1, 1),
            save_file=False,
            numfig=1,
        )

        sys.argv = ["fluidsim-create-xml-description", path_run]
        run()
        sim3.output.temporal_spectra.plot_spectra()
        sim3.output.temporal_spectra.save_data_as_phys_fields(delta_index_times=2)
        sim3.output.temporal_spectra.save_spectra()

        spatiotemporal_spectra = sim3.output.spatiotemporal_spectra
        series_kxkykz = spatiotemporal_spectra.load_time_series()

        spectra_kxkykzomega = spatiotemporal_spectra.compute_spectra()
        spectra_omega_from_spatiotemp = (
            spatiotemporal_spectra.compute_temporal_spectra()
        )
        spectra_omega = sim3.output.temporal_spectra.compute_spectra()

        means = sim3.output.spatial_means.load()

        deltakx = 2 * pi / self.params.oper.Lx
        order = spectra_kxkykzomega["dims_order"]
        KX = deltakx * spectra_kxkykzomega[f"K{order[2]}_adim"]
        kx_max = self.params.oper.nx // 2 * deltakx

        assert kx_max == KX.max()

        from fluidsim.solvers.ns3d.output.spatiotemporal_spectra import (
            _sum_wavenumber3D,
        )

        def sum_wavenumber(field):
            return _sum_wavenumber3D(field, KX, kx_max)

        _ = spatiotemporal_spectra.save_spectra_kzkhomega(save_urud=True)
        spectra_kzkhomega = spatiotemporal_spectra.load_spectra_kzkhomega(
            save_urud=True
        )

        delta_kz = spectra_kzkhomega["kz_spectra"][1]
        delta_kh = spectra_kzkhomega["kh_spectra"][1]
        delta_omega = spectra_kzkhomega["omegas"][1]
        coef = delta_kz * delta_kh * delta_omega

        for letter in "xyz":
            vi_fft = series_kxkykz[f"v{letter}_Fourier"]
            spectrum_kxkykzomega = spectra_kxkykzomega["spectrum_v" + letter]
            spectrum_omega = spectra_omega["spectrum_v" + letter]
            spectrum_omega_from_spatiotemp = spectra_omega_from_spatiotemp[
                "spectrum_v" + letter
            ]
            spectrum_kzkhomega = spectra_kzkhomega["spectrum_v" + letter]

            E_series_kxkykz = 0.5 * sum_wavenumber(
                (abs(vi_fft) ** 2).mean(axis=-1)
            )
            assert E_series_kxkykz > 0, (letter, vi_fft)

            E_kxkykzomega = (
                0.5
                * delta_omega
                * sum_wavenumber(spectrum_kxkykzomega.sum(axis=-1))
            )
            E_omega = 0.5 * delta_omega * spectrum_omega.sum()
            E_omega_from_spatiotemp = (
                0.5 * delta_omega * spectrum_omega_from_spatiotemp.sum()
            )

            E_kzkhomega = 0.5 * coef * spectrum_kzkhomega.sum()

            # `:-1` because the last time is saved twice in spatial_means
            # (see SpatialMeansBase.__init__)
            E_mean = means["E" + letter][:-1].mean()

            assert np.allclose(E_omega, E_kxkykzomega), (
                letter,
                E_kxkykzomega / E_mean,
            )

            assert np.allclose(E_series_kxkykz, E_kxkykzomega), (
                letter,
                E_kxkykzomega / E_series_kxkykz,
            )

            assert np.allclose(E_series_kxkykz, E_kzkhomega), (
                letter,
                E_kzkhomega / E_series_kxkykz,
            )

            assert np.allclose(E_mean, E_series_kxkykz), (
                letter,
                E_series_kxkykz / E_mean,
            )

            assert np.allclose(E_omega, E_omega_from_spatiotemp), (
                letter,
                E_omega,
                E_omega_from_spatiotemp,
            )

            # print(f"{spectrum_omega.sum() / spectrum_omega_from_spatiotemp.sum() = }")

            # this condition is not exactly fulfilled (why?)
            # assert np.allclose(
            #     spectrum_omega, spectrum_omega_from_spatiotemp
            # ), (
            #     letter,
            #     spectrum_omega,
            #     spectrum_omega_from_spatiotemp,
            #     spectrum_omega / spectrum_omega_from_spatiotemp,
            #     spectrum_omega.sum() / spectrum_omega_from_spatiotemp.sum(),
            # )

        spectrum_Khd = spectra_kzkhomega["spectrum_Khd"]
        spectrum_vz = spectra_kzkhomega["spectrum_vz"]

        # because k \cdot \hat v = 0, for kz = 0, Khd = 0
        assert np.allclose(spectrum_Khd[0].sum(), 0.0)

        # because k \cdot \hat v = 0, for kh = 0, Kz = 0
        # `1:` because the energy in the mode kx=ky=kz=0 is not zero.
        assert np.allclose(spectrum_vz[1:, 0, :].sum(), 0.0)

        sim3.output.spatiotemporal_spectra.plot_temporal_spectra()
        sim3.output.spatiotemporal_spectra.plot_kzkhomega(
            key_field="Khr", equation="kh=1"
        )

        plt.close("all")


class TestInitInScript(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.output.HAS_TO_SAVE = False
        params.init_fields.type = "in_script"

    def test_init_in_script(self):

        sim = self.sim

        # here we have to initialize the flow fields

        variables = {
            k: 1e-2 * sim.oper.create_arrayX_random() for k in ("vx", "vy", "vz")
        }

        X, Y, Z = sim.oper.get_XYZ_loc()

        sim.state.init_statephys_from(**variables)

        sim.state.statespect_from_statephys()
        sim.state.statephys_from_statespect()

        vx_fft = sim.state.get_var("vx_fft")
        vy_fft = sim.state.get_var("vy_fft")
        vz_fft = sim.state.get_var("vz_fft")

        sim.state.init_from_vxvyfft(vx_fft, vy_fft)
        sim.state.init_from_vxvyvzfft(vx_fft, vy_fft, vz_fft)


class TestForcingWatuCoriolis(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()

        params.init_fields.type = "noise"

        omega_f = 0.3  # rad/s
        delta_omega_f = 0.03  # rad/s
        N = 0.4  # rad/s
        amplitude = 0.05  # m
        period_N = 2 * np.pi / N
        period_forcing = 1e1 * period_N

        params.forcing.enable = True
        params.forcing.type = "watu_coriolis"

        watu = params.forcing.watu_coriolis
        watu.omega_f = omega_f
        watu.delta_omega_f = delta_omega_f
        watu.amplitude = amplitude
        watu.period_forcing = period_forcing
        watu.approximate_dt = period_N / 1e1
        watu.nb_wave_makers = 2

        params.output.periods_save.phys_fields = 2.0

    def test_forcing(self):

        sim = self.sim
        sim.time_stepping.start()
        params, Simul = load_for_restart(sim.output.path_run)
        params.time_stepping.t_end += 2.0
        sim_restart = Simul(params)
        sim_restart.time_stepping.start()

        if mpi.nb_proc > 1:
            return

        modif_resolution_from_dir(
            self.sim.output.path_run, coef_modif_resol=3.0 / 2, PLOT=True
        )

        path_dir_big = next(Path(self.sim.output.path_run).glob("State_phys_*"))
        sim_big = load_state_phys_file(path_dir_big)

        for key in self.sim.state.keys_state_phys:
            var = sim_restart.state.get_var(key)
            var_big = sim_big.state.get_var(key)
            assert np.allclose(np.mean(var ** 2), np.mean(var_big ** 2))


class TestForcingMilestone(TestSimulBase):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.forcing.enable = True
        params.forcing.type = "milestone"
        movement = params.forcing.milestone.movement
        movement.type = "uniform"
        movement.uniform.speed = 1.0

        params.init_fields.type = "noise"
        params.init_fields.noise.velo_max = 5e-3

        return params

    def test_milestone(self):
        self.sim.time_stepping.start()


class TestForcingMilestonePeriodicUniform(TestForcingMilestone):
    @classmethod
    def init_params(self):
        params = super().init_params()
        params.oper.NO_SHEAR_MODES = True
        params.time_stepping.t_end = 2.0
        params.forcing.milestone.nx_max = 24
        movement = params.forcing.milestone.movement
        movement.type = "periodic_uniform"
        movement.periodic_uniform.length = 2.0
        movement.periodic_uniform.length_acc = 0.25
        movement.periodic_uniform.speed = 2.5

    def test_milestone(self):
        super().test_milestone()
        self.sim.forcing.get_info()


if __name__ == "__main__":
    unittest.main()



"""Utilities for FluidSim
=========================

Provides:

.. autosummary::
   :toctree:

   util
   testing
   console
   mini_oper_modif_resol

"""

from .util import times_start_end_from_path



# -*- coding: utf-8 -*-

"""Stratified NS3D solver (:mod:`fluidsim.solvers.ns3d.strat.solver`)
=====================================================================

.. autoclass:: InfoSolverNS3DStrat
   :members:
   :private-members:

.. autoclass:: Simul
   :members:
   :private-members:

"""

from transonic import boost

from fluidfft.fft3d.operators import vector_product

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.operators.operators3d import dealiasing_variable

from ..solver import InfoSolverNS3D, Simul as SimulNS3D


Ac = "complex128[:,:,:]"


@boost
def compute_fb_fft(div_vb_fft: Ac, N: float, vz_fft: Ac):
    fb_fft = div_vb_fft
    fb_fft[:] = -div_vb_fft - N ** 2 * vz_fft
    return fb_fft


class InfoSolverNS3DStrat(InfoSolverNS3D):
    def _init_root(self):

        super()._init_root()

        package = "fluidsim.solvers.ns3d.strat"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "ns3d.strat"

        classes = self.classes

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StateNS3DStrat"

        # classes.InitFields.module_name = package + '.init_fields'
        # classes.InitFields.class_name = 'InitFieldsNS3D'

        classes.Output.module_name = package + ".output"
        classes.Output.class_name = "Output"


# classes.Forcing.module_name = package + '.forcing'
# classes.Forcing.class_name = 'ForcingNS3D'


class Simul(SimulNS3D):
    r"""Pseudo-spectral solver 3D incompressible Navier-Stokes equations.

    Notes
    -----

    .. |p| mathmacro:: \partial

    .. |vv| mathmacro:: \textbf{v}

    .. |kk| mathmacro:: \textbf{k}

    .. |ek| mathmacro:: \textbf{e}_\textbf{k}

    .. |ez| mathmacro:: \textbf{e}_\textbf{z}

    .. |bnabla| mathmacro:: \boldsymbol{\nabla}

    .. |bomega| mathmacro:: \boldsymbol{\omega}

    This class is dedicated to solve with a pseudo-spectral method the
    incompressible Navier-Stokes equations (possibly with hyper-viscosity)
    under the Boussinesq approximation with a constant Brunt-Vaisala frequency:

    .. math::

      \p_t \vv + \vv \cdot \bnabla \vv =
      - \bnabla p + b \ez - \nu_\alpha (-\Delta)^{\alpha/2} \vv,

      \p_t b   + \vv \cdot \bnabla b =
      - N^2 v_z  - \nu_\alpha (-\Delta)^{\alpha/2} b,

    where :math:`\vv` is the non-divergent velocity (:math:`\bnabla \cdot \vv =
    0`), :math:`p` is the pressure, :math:`\Delta` is the 3D Laplacian
    operator, :math:`b` is the buoyancy and :math:`N` is the (constant)
    Brunt-Vaisala frequency.

    The equation for the velocity can be rewritten as (here without the viscous term)

    .. math::

      \p_t \vv =
      - \bnabla (p + |\vv|^2/2) + \vv \times \bomega  + b \ez,

    In Fourier space, we obtain

    .. math::
      \p_t \hat{\vv} = \hat N(\vv) + L \hat{\vv},

    where

    .. math::

      \hat N(\vv) = P_\perp \widehat{\vv \times \bomega} + \hat b \ez,

    .. math::

      L = - \nu_\alpha |\kk|^{2\alpha},

    with :math:`P_\perp = (1 - \ek \ek \cdot)` the operator projection on the
    plane perpendicular to the wave number :math:`\kk`.

    """
    InfoSolver = InfoSolverNS3DStrat

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        SimulNS3D._complete_params_with_default(params)
        attribs = {"N": 1.0}
        params._set_attribs(attribs)

    @classmethod
    def _modify_sim_repr_maker(cls, sim_repr_maker):
        sim_repr_maker.add_parameters({"N": sim_repr_maker.sim.params.N})

    def tendencies_nonlin(self, state_spect=None, old=None):
        oper = self.oper
        ifft_as_arg = oper.ifft_as_arg
        ifft_as_arg_destroy = oper.ifft_as_arg_destroy
        fft_as_arg = oper.fft_as_arg

        if state_spect is None:
            spect_get_var = self.state.state_spect.get_var
        else:
            spect_get_var = state_spect.get_var

        vx_fft = spect_get_var("vx_fft")
        vy_fft = spect_get_var("vy_fft")
        vz_fft = spect_get_var("vz_fft")
        b_fft = spect_get_var("b_fft")

        omegax_fft, omegay_fft, omegaz_fft = oper.rotfft_from_vecfft(
            vx_fft, vy_fft, vz_fft
        )

        if self.params.f is not None:
            self._modif_omegafft_with_f(omegax_fft, omegay_fft, omegaz_fft)

        omegax = self.state.fields_tmp[3]
        omegay = self.state.fields_tmp[4]
        omegaz = self.state.fields_tmp[5]

        ifft_as_arg_destroy(omegax_fft, omegax)
        ifft_as_arg_destroy(omegay_fft, omegay)
        ifft_as_arg_destroy(omegaz_fft, omegaz)

        if state_spect is None:
            vx = self.state.state_phys.get_var("vx")
            vy = self.state.state_phys.get_var("vy")
            vz = self.state.state_phys.get_var("vz")
        else:
            vx = self.state.fields_tmp[0]
            vy = self.state.fields_tmp[1]
            vz = self.state.fields_tmp[2]
            ifft_as_arg(vx_fft, vx)
            ifft_as_arg(vy_fft, vy)
            ifft_as_arg(vz_fft, vz)

        fx, fy, fz = vector_product(vx, vy, vz, omegax, omegay, omegaz)

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        fx_fft = tendencies_fft.get_var("vx_fft")
        fy_fft = tendencies_fft.get_var("vy_fft")
        fz_fft = tendencies_fft.get_var("vz_fft")

        fft_as_arg(fx, fx_fft)
        fft_as_arg(fy, fy_fft)
        fft_as_arg(fz, fz_fft)

        fz_fft += b_fft

        oper.project_perpk3d(fx_fft, fy_fft, fz_fft)

        if state_spect is None:
            b = self.state.state_phys.get_var("b")
        else:
            b = self.state.fields_tmp[3]
            ifft_as_arg(b_fft, b)

        div_vb_fft = oper.div_vb_fft_from_vb(vx, vy, vz, b)
        fb_fft = compute_fb_fft(div_vb_fft, self.params.N, vz_fft)

        if self.no_vz_kz0:
            dealiasing_variable(fz_fft, self.where_kz_0)
            dealiasing_variable(fb_fft, self.where_kz_0)

        tendencies_fft.set_var("b_fft", fb_fft)

        if self.is_forcing_enabled:
            tendencies_fft += self.forcing.get_forcing()

        self.oper.dealiasing(tendencies_fft)
        return tendencies_fft


if __name__ == "__main__":

    import numpy as np

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    n = 16
    L = 2 * np.pi
    params.oper.nx = n
    params.oper.ny = n
    params.oper.nz = n
    params.oper.Lx = L
    params.oper.Ly = L
    params.oper.Lz = L
    params.oper.type_fft = "fluidfft.fft3d.mpi_with_fftwmpi3d"
    # params.oper.type_fft = 'fluidfft.fft3d.with_pyfftw'
    # params.oper.type_fft = 'fluidfft.fft3d.with_cufft'

    delta_x = params.oper.Lx / params.oper.nx
    # params.nu_8 = 2.*10e-1*params.forcing.forcing_rate**(1./3)*delta_x**8
    params.nu_8 = 2.0 * 10e-1 * delta_x ** 8

    params.time_stepping.USE_T_END = True
    params.time_stepping.t_end = 6.0
    params.time_stepping.it_end = 2

    params.init_fields.type = "noise"
    params.init_fields.noise.velo_max = 1.0
    params.init_fields.noise.length = 1.0

    # params.forcing.enable = False
    # params.forcing.type = 'random'
    # 'Proportional'
    # params.forcing.type_normalize

    params.output.periods_print.print_stdout = 0.00000000001

    params.output.periods_save.phys_fields = 1.0
    # params.output.periods_save.spectra = 0.5
    # params.output.periods_save.spatial_means = 0.05
    # params.output.periods_save.spect_energy_budg = 0.5

    # params.output.periods_plot.phys_fields = 0.0

    params.output.ONLINE_PLOT_OK = True

    # params.output.spectra.HAS_TO_PLOT_SAVED = True
    # params.output.spatial_means.HAS_TO_PLOT_SAVED = True
    # params.output.spect_energy_budg.HAS_TO_PLOT_SAVED = True

    # params.output.phys_fields.field_to_plot = 'rot'

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    # sim.output.phys_fields.plot()

    fld.show()



"""Simple 1d solver (:mod:`fluidsim.solvers.square1d.solver`)
=============================================================

Provides:

.. autoclass:: Simul
   :members:
   :private-members:

"""
import numpy as np

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)


class InfoSolverSquare1DPseudoSpect(InfoSolverPseudoSpectral):
    def _init_root(self):
        super()._init_root()

        package = "fluidsim.solvers.nl1d"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "nl1d"

        classes = self.classes

        classes.Operators.module_name = "fluidsim.operators.operators1d"
        classes.Operators.class_name = "OperatorsPseudoSpectral1D"

        classes.State.module_name = "fluidsim.solvers.ad1d.pseudo_spect.state"
        classes.State.class_name = "StateAD1DPseudoSpectral"

        base = "fluidsim.solvers.ad1d"

        classes.Output.module_name = base + ".output"
        classes.Output.class_name = "Output"

        classes.InitFields.module_name = base + ".init_fields"
        classes.InitFields.class_name = "InitFieldsAD1D"


class Simul(SimulBasePseudoSpectral):
    InfoSolver = InfoSolverSquare1DPseudoSpect

    @staticmethod
    def _complete_params_with_default(params):
        SimulBasePseudoSpectral._complete_params_with_default(params)
        params._set_attrib("sigma", 1.0)

    def tendencies_nonlin(self, state_spect=None, old=None):

        if state_spect is None:
            signal = self.state.state_phys.get_var("s")
        else:
            s_fft = state_spect.get_var("s_fft")
            signal = self.oper.ifft(s_fft)

        f_signal = -np.sign(signal) * self.params.sigma * signal ** 2

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old
        f_fft = tendencies_fft.get_var("s_fft")
        self.oper.fft_as_arg(f_signal, f_fft)
        self.oper.dealiasing(tendencies_fft)
        return tendencies_fft


if __name__ == "__main__":

    from scipy.signal import gausspulse

    params = Simul.create_default_params()
    params.output.sub_directory = "examples"

    params.output.periods_save.phys_fields = 1.0
    params.short_name_type_run = "test"
    params.oper.Lx = 10
    params.oper.nx = 256
    params.nu_2 = 0
    params.time_stepping.USE_CFL = False
    params.time_stepping.deltat0 = 1e-3
    params.time_stepping.t_end = 10.0
    params.time_stepping.type_time_scheme = "RK2"
    params.init_fields.type = "in_script"

    sim = Simul(params)

    # Initialize
    x = sim.oper.x
    Lx = sim.oper.Lx
    s = gausspulse(x - Lx / 2, fc=2)
    s_fft = sim.oper.fft(s)

    sim.state.init_statephys_from(s=s)
    sim.state.init_statespect_from(s_fft=s_fft)

    # Plot once
    sim.output.init_with_initialized_state()
    sim.output.phys_fields.plot(field="s")

    sim.time_stepping.start()
    sim.output.phys_fields.plot(field="s")



import unittest

import fluiddyn.util.mpi as mpi
from fluidsim.util.testing import (
    TestSimulConserve,
    classproperty,
    skip_if_no_fluidfft,
)


@skip_if_no_fluidfft
class TestSimulSW1LExactlinModified(TestSimulConserve):
    @classproperty
    def Simul(cls):
        from fluidsim.solvers.sw1l.exactlin.modified.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):
        super().init_params()
        params = cls.params
        params.output.HAS_TO_SAVE = True

        params.output.periods_save.spect_energy_budg = 0.2

        params.forcing.enable = True
        params.forcing.type = "potential"
        params.forcing.nkmin_forcing = 2
        params.forcing.nkmax_forcing = 4
        params.init_fields.type = "vortex_grid"
        params.oper.nx = 16
        params.oper.ny = 8

    def get_tendencies(self):
        tendencies_fft = self.tendencies_fft

        Fq_fft = tendencies_fft.get_var("q_fft")
        Fap_fft = tendencies_fft.get_var("ap_fft")
        Fam_fft = tendencies_fft.get_var("am_fft")

        return self.sim.oper.uxuyetafft_from_qapamfft(Fq_fft, Fap_fft, Fam_fft)

    # def test_enstrophy_conservation(self):
    #     """Theoretically not expected to conserve quadratic potential enstrophy"""
    #     pass

    def test_energy_conservation(self):
        """Verify that the energy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state_phys = self.sim.state.state_phys
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        oper = self.sim.oper
        Fx = oper.ifft2(Fx_fft)
        Fy = oper.ifft2(Fy_fft)
        Feta = oper.ifft2(Feta_fft)
        A = (ux * Fx + uy * Fy) + self.sim.params.c2 * eta * Feta

        A_fft = oper.fft2(A)
        if mpi.rank == 0:
            self.assertAlmostZero(A_fft[0, 0])

    def test_energy_conservation_fft(self):
        """Verify that the quadratic energy growth rate due to nonlinear
        tendencies (advection term) must be zero in the spectral plane.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state = self.sim.state
        ux_fft = state.get_var("ux_fft")
        uy_fft = state.get_var("uy_fft")
        eta_fft = state.get_var("eta_fft")

        oper = self.sim.oper
        T_ux = (ux_fft.conj() * Fx_fft).real
        T_uy = (uy_fft.conj() * Fy_fft).real
        T_eta = (eta_fft.conj() * Feta_fft).real * self.sim.params.c2
        T_tot = T_ux + T_uy + T_eta
        sum_T = oper.sum_wavenumbers(T_tot)
        self.assertAlmostZero(sum_T)


if __name__ == "__main__":
    unittest.main()



"""State of the variables (:mod:`fluidsim.base.basilisk.state`)
===============================================================

Provides:

.. autoclass:: StateBasilisk
   :members:
   :private-members:

"""


from ..state import StateBase


class StateBasilisk(StateBase):
    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_phys": ["ux", "uy", "rot"],
                "keys_computable": [],
                "keys_phys_needed": ["rot"],
            }
        )

    def compute(self, key):
        """Compute scalar fields such a component of the velocity or vorticity."""
        if key == "ux":
            scalar = self.sim.basilisk.uf.x
        elif key == "uy":
            scalar = self.sim.basilisk.uf.y
        elif key == "rot":
            scalar = self.sim.basilisk.omega
        else:
            raise ValueError('No method to compute key "' + key + '"')

        return scalar.f(self.oper.X, self.oper.Y)

    def _get_state_from_basilisk(self):
        for key in self.keys_state_phys:
            self.state_phys.set_var(key, self.compute(key))



"""Operators 0d (:mod:`fluidsim.operators.operators0d`)
=======================================================

Provides

.. autoclass:: Operators0D
   :members:
   :private-members:

"""

from fluiddyn.util import mpi


class Operators0D:
    """0D operators."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        params._set_child("oper")

    def __init__(self, params=None, SEQUENTIAL=None):
        if mpi.nb_proc > 1:
            raise ValueError

        self.params = params
        self.axes = tuple()
        self.shapeX_seq = self.shapeX_loc = []

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""

        return ""

    def produce_long_str_describing_oper(self):
        """Produce a string describing the operator."""

        return "0d simulation\n"

    def gather_Xspace(self, a):
        """Gather an array (mpi), in this case, just return the array."""
        return a



"""
FrequencySpectra (:mod:`fluidsim.solvers.ns2d.strat.output.frequency_spectra`)
==============================================================================


Provides:

.. autoclass:: FrequencySpectra
   :members:
   :private-members:

"""

import sys
import time
import h5py
import numpy as np

from scipy import signal
from pathlib import Path

from fluiddyn.util import mpi
from fluidsim.base.output.base import SpecificOutput


class FrequencySpectra(SpecificOutput):
    """
    Computes the frequency spectra.
    """

    _tag = "frequency_spectra"
    _name_file = _tag + ".h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "frequency_spectra"

        params.output.periods_save._set_attrib(tag, 0)
        params.output._set_child(
            tag,
            attribs={
                "HAS_TO_PLOT_SAVED": False,
                "time_start": 1,
                "time_decimate": 1,
                "spatial_decimate": 2,
                "size_max_file": 1e-2,
            },
        )

    def __init__(self, output):
        params = output.sim.params
        pfreq_spectra = params.output.frequency_spectra
        super().__init__(
            output,
            period_save=params.output.periods_save.frequency_spectra,
            has_to_plot_saved=pfreq_spectra.HAS_TO_PLOT_SAVED,
        )

        # Parameters
        self.time_start = pfreq_spectra.time_start
        self.time_decimate = pfreq_spectra.time_decimate
        self.spatial_decimate = pfreq_spectra.spatial_decimate
        self.size_max_file = pfreq_spectra.size_max_file
        self.periods_save = params.output.periods_save.frequency_spectra

        self.has_to_save = bool(params.output.periods_save.frequency_spectra)

        # Check: In restart... time_start == time last state.
        path_run = Path(self.sim.output.path_run)
        path_spatio_temp_files = path_run / "frequency_spectra"

        if sorted(path_spatio_temp_files.glob("temp_array_it*")):
            name = sorted(path_run.glob("state_phys*"))[-1].stem
            if "_it" in name:
                time_last_file = float(
                    name.split("state_phys_t")[-1].split("_it")[0]
                )
            else:
                time_last_file = float(name.split("state_phys_t")[1])

            if round(self.time_start, 3) != time_last_file:
                self.time_start = time_last_file

        # Number points each direction
        n0 = len(list(range(0, params.oper.ny, self.spatial_decimate)))
        n1 = len(list(range(0, params.oper.nx, self.spatial_decimate)))

        # Compute number array in file
        nb_bytes = np.empty([n0, n1], dtype=float).nbytes
        self.nb_arr_in_file = int(self.size_max_file * (1024 ** 2) // nb_bytes)
        mpi.printby0("nb_arr_in_file", self.nb_arr_in_file)

        # Check: duration file <= duration simulation
        self.duration_file = (
            self.nb_arr_in_file
            * self.params.time_stepping.deltat0
            * self.time_decimate
        )
        if (
            self.duration_file > self.params.time_stepping.t_end
            and self.periods_save > 0
        ):
            raise ValueError(
                "The duration of the simulation is not enough to fill a file."
            )

        # Check: self.nb_arr_in_file should be > 0
        if self.nb_arr_in_file <= 0 and self.periods_save > 0:
            raise ValueError("The size of the file should be larger.")

        else:
            self.temp_array = np.empty([self.nb_arr_in_file, n0, n1], dtype=float)

            # Array 4D (2 keys, times, n0, n1)
            self.temp_array_new = np.empty(
                [2, self.nb_arr_in_file, n0, n1], dtype=float
            )

        # Convert time_start to it_start.
        if not self.sim.time_stepping.it:
            self.it_start = int(
                self.time_start / self.params.time_stepping.deltat0
            )
        else:
            # If simulation starts from an specific time...
            self.it_start = self.sim.time_stepping.it + int(
                (self.time_start - self.sim.time_stepping.t)
                / self.params.time_stepping.deltat0
            )

            if self.it_start < self.sim.time_stepping.it:
                self.it_start = self.sim.time_stepping.it

        # Create empty array with times
        self.times_arr = np.empty([self.nb_arr_in_file])

        if (
            params.time_stepping.USE_CFL
            and params.output.periods_save.frequency_spectra > 0
        ):
            raise ValueError(
                "To compute the frequency spectra: \n"
                + "USE_CFL = FALSE and periods_save.frequency_spectra > 0"
            )

        # Create directory to save files
        if mpi.rank == 0:
            dir_name = self._tag
            self.path_dir = Path(self.sim.output.path_run) / dir_name

        if self.has_to_save and mpi.rank == 0:
            self.path_dir.mkdir(exist_ok=True)

        # Start loop in _online_save
        self.it_last_run = self.it_start
        self.nb_times_in_temp_array = 0

    def _init_files(self, arrays_1st_time=None):
        # we can not do anything when this function is called.
        pass

    def _write_to_file(self, temp_arr, times_arr):
        """Writes a file with the temporal data"""
        if mpi.rank == 0:
            # Name file
            it_start = int(times_arr[0] / self.sim.params.time_stepping.deltat0)
            name_file = f"temp_array_it{it_start}.h5"
            path_file = self.path_dir / name_file

            # Dictionary arrays
            dict_arr = {
                "it_start": it_start,
                "times_arr": times_arr,
                "temp_arr": temp_arr,
            }

            # Write dictionary to file
            with h5py.File(path_file, "w") as file:
                for k, v in list(dict_arr.items()):
                    file.create_dataset(k, data=v)

    def _online_save(self):
        """Computes and saves the values at one time."""
        oper = self.sim.oper
        if self.periods_save == 0:
            pass
        else:
            itsim = self.sim.time_stepping.it
            if itsim - self.it_last_run >= self.time_decimate:
                self.it_last_run = itsim

                # Compute linear mode in physical space
                field_ap = self.sim.state.compute("ap")
                field_am = self.sim.state.compute("am")

                if mpi.nb_proc > 1:
                    # Create big array
                    if mpi.rank == 0:
                        field_ap_seq = oper.create_arrayX(shape="seq")
                        field_am_seq = oper.create_arrayX(shape="seq")
                    else:
                        field_ap_seq = None
                        field_am_seq = None

                    # Define size of each array
                    sendcounts = np.array(mpi.comm.gather(field_ap.size, root=0))

                    # Send array each process to root (process 0)
                    mpi.comm.Gatherv(
                        sendbuf=field_ap,
                        recvbuf=(field_ap_seq, sendcounts),
                        root=0,
                    )
                    mpi.comm.Gatherv(
                        sendbuf=field_am,
                        recvbuf=(field_am_seq, sendcounts),
                        root=0,
                    )

                else:
                    field_ap_seq = field_ap
                    field_am_seq = field_am

                # Decimate with process 0
                if mpi.rank == 0:
                    field_ap_decimate = field_ap_seq[
                        :: self.spatial_decimate, :: self.spatial_decimate
                    ]
                    field_am_decimate = field_am_seq[
                        :: self.spatial_decimate, :: self.spatial_decimate
                    ]

                    self.temp_array_new[
                        0, self.nb_times_in_temp_array, :, :
                    ] = field_ap_decimate
                    self.temp_array_new[
                        1, self.nb_times_in_temp_array, :, :
                    ] = field_am_decimate

                # Save the time to self.times_arr
                self.times_arr[self.nb_times_in_temp_array] = (
                    itsim * self.sim.params.time_stepping.deltat0
                )

                # Check if self.temp_array_new is filled. If yes, writes to a file.
                if self.nb_times_in_temp_array == self.nb_arr_in_file - 1:
                    if mpi.rank == 0:
                        print("Saving temporal data...")
                    self._write_to_file(self.temp_array_new, self.times_arr)

                    self.nb_times_in_temp_array = 0
                else:
                    self.nb_times_in_temp_array += 1

    def _concatenate_data(self, tmin, tmax):
        """
        Concatenates temporal data.

        Argument
            tmin {float} -- Minimum time to perform temporal FT.
            tmax {float} -- Maximum time to perform temporal FT.

        Raises:
            ValueError -- itmin == itmax

        Returns:
            times_conc {array} -- Concatenated time array
            temp_arr_conc {array} -- concatenated temporal data array
        """

        path_files = sorted(self.path_dir.glob("temp_array_it*"))

        # Sort files by number of iteration.
        list_its = []

        for path_file in path_files:
            list_its.append(int(path_file.name.split("_it")[1].split(".h5")[0]))

        list_its = sorted(list_its)
        print(list_its)
        list_files_new = []
        for it in list_its:
            list_files_new.append(
                self.path_dir / ("temp_array_it" + str(it) + ".h5")
            )

        list_files = list_files_new
        for lf in list_files:
            print(lf.name)

        # Create concatenated arrays
        temp_arr_conc = None
        times_conc = None

        for path_file in list_files:
            with h5py.File(path_file, "r") as file:
                temp_arr = file["temp_arr"][...]
                times = file["times_arr"][...]

            # Concatenate arrays
            if isinstance(temp_arr_conc, np.ndarray):
                temp_arr_conc = np.concatenate((temp_arr_conc, temp_arr), axis=1)
            if isinstance(times_conc, np.ndarray):
                times_conc = np.concatenate((times_conc, times), axis=0)
            else:
                temp_arr_conc = temp_arr
                times_conc = times

        # Default tmin and tmax
        if not tmin:
            tmin = np.min(times_conc)
        if not tmax:
            tmax = np.max(times_conc)

        # Find indexes itmin and itmax
        itmin = np.argmin(abs(times_conc - tmin))
        itmax = np.argmin(abs(times_conc - tmax)) + 1

        # Check..
        if itmin == itmax:
            raise ValueError(
                "itmin == itmax. \n"
                + f"Choose between tmin = {np.min(times_conc)} - "
                + f"tmax = {np.max(times_conc)}"
            )

        # Print info..
        to_print = f"tmin = {np.min(times_conc)} ; tmax = {np.max(times_conc)}"
        print(to_print)

        return times_conc[itmin:itmax], temp_arr_conc[:, itmin:itmax, :, :]

    def compute_frequency_spectra(self):
        """
        Computes and saves the frequency spectra.
        """
        # Define list of path files
        list_files = sorted(self.path_dir.glob("temp_array_iself.t*"))

        # Compute sampling frequency
        freq_sampling = 1.0 / (
            self.time_decimate * self.params.time_stepping.deltat0
        )

        for index, file_path in enumerate(list_files):

            # Generating counter
            print(
                "Computing frequency spectra = {}/{}".format(
                    index, len(list_files) - 1
                ),
                end="\r",
            )

            # Load data from file
            with h5py.File(file_path, "r") as file:
                temp_array = file["temp_arr"][...]
                times = file["times_arr"][...]

            # Compute the temporal spectrum of a 3D array
            omegas, freq_spectrum = signal.periodogram(
                temp_array,
                fs=freq_sampling,
                window="hann",
                nfft=temp_array.shape[1],
                detrend="constant",
                return_onesided=False,
                scaling="spectrum",
                axis=1,
            )

            # Save array omegas and spectrum to file
            dict_arr = {"omegas": omegas, "freq_spectrum": freq_spectrum}

            with h5py.File(file_path, "r+") as file:
                for k, v in list(dict_arr.items()):
                    file.create_dataset(k, data=v)

            # Flush buffer and sleep time
            sys.stdout.flush()
            time.sleep(0.2)

    def _init_online_plot(self):
        if mpi.rank == 0:
            fig, axe = self.output.figure_axe(numfig=1_000_000)
            self.axe = axe
            axe.set_xlabel("$k_h$")
            axe.set_ylabel("$E(k_h)$")
            axe.set_title("spectra\n" + self.output.summary_simul)
